from __future__ import absolute_import
from __future__ import print_function

import subprocess
import atexit
import os
import socket
from threading import Thread, Lock

import coloredlogs
import logging

import copy
import requests
import signal
import json
import time
import six
import PIL.Image
import sys
import msgpack

from aetros.const import JOB_STATUS
from aetros.git import Git
from aetros.logger import GeneralLogger
from aetros.utils import git, invalid_json_values, read_config

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

from aetros.MonitorThread import MonitoringThread


def on_shutdown():
    for job in on_shutdown.started_jobs:
        job.on_shutdown()


on_shutdown.started_jobs = []

atexit.register(on_shutdown)


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


class EventListener:
    def __init__(self):
        self.events = {}

    def on(self, name, callback):
        if name not in self.events:
            self.events[name] = []

        self.events[name].append(callback)

    def fire(self, name, parameter=None):
        if name in self.events:
            for callback in self.events[name]:
                callback(parameter)


class ApiClient:
    def __init__(self, api_host, api_key):
        self.host = api_host
        self.api_key = api_key

    def get(self, url, params=None, **kwargs):
        json_chunk = kwargs.get('json')
        if (json_chunk and not isinstance(json_chunk, str)):
            kwargs['json'] = json.loads(json.dumps(json_chunk, default=invalid_json_values))

        return requests.get(self.get_url(url), params=params, **kwargs)

    def post(self, url, data=None, **kwargs):
        json_chunk = kwargs.get('json')
        if (json_chunk and not isinstance(json_chunk, str)):
            kwargs['json'] = json.loads(json.dumps(json_chunk, default=invalid_json_values))

        return requests.post(self.get_url(url), data=data, **kwargs)

    def put(self, url, data=None, **kwargs):
        json_chunk = kwargs.get('json')
        if (json_chunk and not isinstance(json_chunk, str)):
            kwargs['json'] = json.loads(json.dumps(json_chunk, default=invalid_json_values))

        return requests.put(self.get_url(url), data=data, **kwargs)

    def get_url(self, affix):

        url = 'http://%s/api/%s' % (self.host, affix)

        if self.api_key:
            if '?' in url:
                url += '&token=' + self.api_key
            else:
                url += '?token=' + self.api_key

        return url


class BackendClient:
    def __init__(self, config, event_listener, logger):
        """
        :type api_host: string
        :type api_key: string
        :type event_listener: EventListener
        :type job_id: integer
        """
        self.host = config['host']
        self.ssh_stream = None
        self.ssh_command = config['ssh']
        self.ssh_key_path = config['ssh_key']

        self.event_listener = event_listener
        self.logger = logger
        self.message_id = 0

        self.api_key = None
        self.job_id = None

        self.thread_read_instance = None
        self.thread_write_instance = None

        self.lock = Lock()
        self.connection_errors = 0
        self.queue = []
        self.connection_tries = 0
        self.in_connecting = False

        # indicates whether we are offline or not, means not connection to the internet and
        # should not establish a connection to Aetros.
        self.online = True

        # Whether the client is active and should do things.
        self.active = False
        self.expect_close = False
        self.external_stopped = False
        self.registered = False
        self.connected = False
        self.was_connected_once = False
        self.read_unpacker = msgpack.Unpacker(encoding='utf-8')

    def start(self):
        self.active = True

        if not self.thread_read_instance:
            self.thread_read_instance = Thread(target=self.thread_read)
            self.thread_read_instance.daemon = True
            self.thread_read_instance.start()

        if not self.thread_write_instance:
            self.thread_write_instance = Thread(target=self.thread_write)
            self.thread_write_instance.daemon = True
            self.thread_write_instance.start()

    def on_connect(self, reconnect=False):
        pass

    def go_offline(self):
        if not self.online:
            return

        self.event_listener.fire('offline')
        self.online = False

    def connect(self):
        """
        In the write-thread we detect that no connection is living anymore and try always again.
        Up to the 3 connection try, we report to user. We keep trying but in silence.
        Also, when more than 10 connection tries are detected, we delay extra 60 seconds.
        """
        if self.connection_tries > 10:
            # We only try in 1 minute again
            time.sleep(60)

        if self.in_connecting:
            return False

        self.in_connecting = True

        self.logger.debug('Wanna connect ...')

        try:
            if self.connected or not self.online:
                return True

            def preexec_fn():
                # Ignore the SIGINT signal by setting the handler to the standard
                # signal handler SIG_IGN.
                signal.signal(signal.SIGINT, signal.SIG_IGN)

            args = [self.ssh_command] if isinstance(self.ssh_command, six.string_types) else self.ssh_command
            args += ['-o', 'StrictHostKeyChecking no']

            if self.ssh_key_path:
                args += ['-i', self.ssh_key_path]

            args = args + ['git@' + self.host, 'stream']
            self.logger.debug('Open ssh: ' + (' '.join(args)))

            self.ssh_stream = subprocess.Popen(args, preexec_fn=preexec_fn, bufsize=0,
                                               stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE)

            self.logger.debug('Done: pid=' + str(self.ssh_stream.pid))
            messages = self.wait_for_at_least_one_message()
            stderrdata = ''

            if not messages:
                stderrdata = self.ssh_stream.stderr.read().decode("utf-8").strip()
            else:
                self.connected = True
                self.registered = self.on_connect(self.was_connected_once)

            if not self.registered:
                self.connected = False
                self.ssh_stream.kill()
                self.connection_tries += 1
                self.go_offline()

                if stderrdata:
                    if 'Connection refused' not in stderrdata and 'Permission denied' not in stderrdata:
                        self.logger.error(stderrdata)

                if 'Permission denied' in stderrdata:
                    if self.connection_tries < 3:
                        self.logger.warning("Access denied. Did you setup your ssh key correctly and saved it in your AETROS Trainer user account?")

                    self.close()
                    sys.exit(1)
            else:
                self.was_connected_once = True

        except Exception as error:
            self.connected = False
            self.registered = False
            if self.connection_tries < 3:
                self.logger.error("Connection error during connecting to %s: %s" % (self.host, str(error)))
            raise
        finally:
            self.in_connecting = False

        return self.connected

    def debug(self):
        sent = len(filter(lambda x: x['_sent'], self.queue))
        sending = len(filter(lambda x: x['_sending'], self.queue))
        open = len(filter(lambda x: not x['_sending'], self.queue))
        self.logger.debug("%d sent, %d in sending, %d open " % (sent, sending, open))

    def end(self):
        self.expect_close = True
        self.send_message({'type': 'end'})
        self.wait_for_close()

    def connection_error(self, error=None):
        if not self.active:
            return

        if self.expect_close:
            return

        if not self.connected:
            return

        # needs to be set before logger.error, since they can call send_message again
        self.connected = False
        self.registered = False

        if socket is None:
            # python interpreter is already dying, so quit
            return

        message = "Connection error"

        if error:
            self.logger.error(message + ": " + str(error))
        else:
            self.logger.error(message)

        import traceback
        self.logger.debug(traceback.format_exc())

        self.event_listener.fire('disconnect')
        self.connection_errors += 1

    def thread_write(self):
        while self.active:
            if self.online:
                if self.connected and self.registered:
                    queue_copy = self.queue[:]

                    try:
                        sent_size = 0
                        sent = []

                        for message in queue_copy:
                            if message['_sending'] and not message['_sent']:
                                message['_sending'] = False

                        for message in queue_copy:
                            if not self.connected or not self.registered:
                                # additional check to make sure there's no race condition
                                break

                            if not message['_sending'] and not message['_sent']:
                                size = self.send_message(message)
                                if size is not False:
                                    sent.append(message)

                                    sent_size += size
                                    if sent_size > 1024 * 1024:
                                        # not too much at once (max 1MB), so we have time to listen for incoming messages
                                        break

                        self.lock.acquire()
                        for message in sent:
                            if message in self.queue:
                                self.queue.remove(message)
                        self.lock.release()
                    except SystemExit:
                        return
                    except KeyboardInterrupt:
                        return
                    finally:
                        if self.lock.locked():
                            self.lock.release()

                if not self.connected and not self.expect_close:
                    if not self.connect():
                        time.sleep(5)

            time.sleep(0.01)

    def thread_read(self):
        while self.active:
            if self.online:
                if self.connected and self.registered:
                    try:
                        messages = self.read() # this blocks

                        if messages is not None:
                            self.handle_messages(messages)

                        continue
                    except SystemExit:
                        return
                    except KeyboardInterrupt:
                        return
                    except:
                        pass

            time.sleep(0.01)

    def wait_sending_last_messages(self):
        # send all missing messages
        i = 0
        while self.registered:
            if len(self.queue) == 0:
                break

            i += 1
            time.sleep(0.1)
            if i % 50 == 0:
                self.logger.warning("We still sync job data. %d messages left. " % (len(self.queue),))

    def wait_for_close(self):
        self.active = False

        if not self.online:
            return

        if not self.ssh_stream:
            return

        while self.ssh_stream.poll() is None:
            time.sleep(0.1)

        self.online = False

    def close(self):
        self.active = False
        self.connected = False

        self.event_listener.fire('close')

        if self.ssh_stream:
            self.ssh_stream.kill()

        self.online = False

    def send(self, message):
        if not self.active:
            return
            # raise Exception("Requested to send messages, but client is inactive.")

        self.message_id += 1
        message['_id'] = self.message_id
        message['_sending'] = False
        message['_sent'] = False

        self.queue.append(message)

    def send_message(self, message):
        if not self.connected:
            return False

        message['_sending'] = True

        import msgpack
        msg = msgpack.packb(message, default=invalid_json_values)

        try:
            self.ssh_stream.stdin.write(msg)
            message['sent'] = True
            self.ssh_stream.stdin.flush()

            return len(msg)
        except KeyboardInterrupt:
            if message['sent']:
                return len(msg)

            return False

        except Exception as error:
            self.connection_error(error)
            return False

    def handle_messages(self, messages):
        for message in messages:
            if not self.external_stopped and 'stop' == message['a']:
                self.external_stopped = True
                self.event_listener.fire('stop', message['force'])

    def wait_for_at_least_one_message(self):
        """
        Reads until we receive at least one message we can unpack. Return all found messages.
        """

        unpacker = msgpack.Unpacker(encoding='utf-8')

        while True:
            chunk = ''
            try:
                chunk = self.ssh_stream.stdout.read(1)
            except Exception as error:
                self.connection_error(error)
                raise

            if chunk == '':
                # happens only when connection broke. If nothing is to be received, it hangs instead.
                self.connection_error('Connection broken')
                return False

            unpacker.feed(chunk)

            messages = [m for m in unpacker]
            if messages:
                return messages

    def read(self):
        """
        Reads from the socket and tries to unpack the message. If successful (because msgpack was able to unpack)
        then we return that message. Else None. Keep calling .read() when new data is available so we try it 
        again.
        """

        try:
            chunk = self.ssh_stream.stdout.read(1)
        except Exception as error:
            self.connection_error(error)
            raise

        if chunk == '':
            # socket connection broken
            self.connection_error('Connection broken')
            return None

        # self.read_buffer.seek(0, 2) #make sure we write at the end
        self.read_unpacker.feed(chunk)

        # self.read_buffer.seek(0)
        messages = [m for m in self.read_unpacker]

        return messages if messages else None


class JobClient(BackendClient):
    def configure(self, model_name, job_id):
        self.model_name = model_name
        self.job_id = job_id

    def on_connect(self, reconnect=False):
        self.send_message({'type': 'register_job_worker', 'model': self.model_name, 'job': self.job_id, 'reconnect': reconnect})
        self.logger.debug("Wait for one client registration")
        messages = self.wait_for_at_least_one_message()
        self.logger.debug("Got " + str(messages))

        if not messages:
            self.event_listener.fire('registration_failed', {'reason': 'No answer received.'})
            return False

        message = messages.pop(0)
        self.logger.debug("handle message: " + str(message))
        if isinstance(message, dict) and 'a' in message:
            if 'aborted' == message['a']:
                self.logger.error("Job aborted meanwhile. Exiting")
                self.event_listener.fire('stop', False)
                self.active = False
                return False

            if 'registration_failed' == message['a']:
                self.event_listener.fire('registration_failed', {'reason': message['reason']})
                return False

            if 'registered' == message['a']:
                self.registered = True
                self.event_listener.fire('registration')
                self.handle_messages(messages)
                return True

        self.logger.error("Registration of job %s failed." % (self.job_id,))
        return False


def start_job(name=None):
    """
    Tries to load the job defined in the AETROS_JOB_ID environment variable.
    If not defined, it creates a new job.
    Starts the job as well.

    :param name: string: model name
    :return: JobBackend
    """

    job = JobBackend(name)
    job.setup_std_output_logging()

    if os.getenv('AETROS_JOB_ID'):
        job.load(os.getenv('AETROS_JOB_ID'))
    else:
        job.create()

    job.start()

    return job


def create_job(name):
    """
    Creates a new job.

    :param name: string : model name
    :return: JobBackend
    """
    job = JobBackend(name)
    job.create()

    return job


class JobLossChannel:
    """
    :type job_backend : JobBackend
    """

    def __init__(self, job_backend, name, xaxis=None, yaxis=None, layout=None):
        self.name = name
        self.job_backend = job_backend
        message = {
            'name': self.name,
            'traces': [{'name': 'training'}, {'name': 'validation'}],
            'type': JobChannel.NUMBER,
            'main': True,
            'xaxis': xaxis,
            'yaxis': yaxis,
            'layout': layout,
            'lossChannel': True
        }

        self.job_backend.git.commit_json_file('CREATE_CHANNEL', 'aetros/job/channel/' + name+ '/config', message)
        self.stream = self.job_backend.git.stream_file('aetros/job/channel/' + name+ '/data.csv')
        self.stream.write('"x","training","validation"\n')

    def send(self, x, training_loss, validation_loss):
        line = json.dumps([x, training_loss, validation_loss])[1:-1]
        self.stream.write(line + "\n")
        self.job_backend.git.store_file('aetros/job/channel/' + self.name + '/last.csv', line)


class JobImage:
    def __init__(self, id, image, label=None):
        self.id = id
        if not isinstance(image, PIL.Image.Image):
            raise Exception('JobImage requires a PIL.Image as image argument.')

        self.image = image
        self.label = label


class JobChannel:
    NUMBER = 'number'
    TEXT = 'text'

    """
    :type job_backend: JobBackend
    """

    def __init__(self, job_backend, name, traces=None,
                 main=False, kpi=False, kpiTrace=0, max_optimization=True,
                 type=None, xaxis=None, yaxis=None, layout=None):
        """
        :param job_backend: JobBakend
        :param name: str
        :param traces: None|list : per default create a trace based on "name".
        :param main: bool : whether this channel is visible in the job list as column for better comparison.

        :param kpi: bool : whether this channel is the KPI (key performance indicator).
                           Used for hyperparameter optimization. Only one channel can be a kpi. Only first trace used.
        :param kpiTrace: bool : if you have multiple traces, define which is the KPI. 0 based index.

        :param max_optimization: bool : whether the optimization maximizes or minmizes the kpi . Use max_optimization=False to
                                        tell the optimization algorithm that his channel minimizes a kpi, for instance the loss of a model.

        :param type: str : One of JobChannel.NUMBER, JobChannel.TEXT, JobChannel.IMAGE
        :param xaxis: dict
        :param yaxis: dict
        :param layout: dict
        """
        self.name = name
        self.job_backend = job_backend
        self.kpi = kpi
        self.kpiTrace = kpiTrace

        if self.kpi:
            self.job_backend.kpi_channel = self

        if not (isinstance(traces, list) or traces is None):
            raise Exception(
                'traces can only be None or a list of dicts: [{name: "name", option1: ...}, {name: "name2"}, ...]')

        if not traces:
            traces = [{'name': name}]

        if isinstance(traces, list) and isinstance(traces[0], six.string_types):
            traces = list(map(lambda x: {'name': x}, traces))

        message = {
            'name': name,
            'traces': traces,
            'type': type or JobChannel.NUMBER,
            'main': main,
            'kpi': kpi,
            'kpiTrace': kpiTrace,
            'maxOptimization': max_optimization,
            'xaxis': xaxis,
            'yaxis': yaxis,
            'layout': layout,
        }
        self.traces = traces
        self.job_backend.git.commit_json_file('CREATE_CHANNEL', 'aetros/job/channel/' + name+ '/config', message)
        self.stream = self.job_backend.git.stream_file('aetros/job/channel/' + name+ '/data.csv')

        if self.kpi:
            self.job_backend.git.commit_file('KPI_CHANNEL', 'aetros/job/kpi/name', name)

        line = json.dumps(['x'] + [str(x['name']) for x in traces])[1:-1]
        self.stream.write(line + "\n")

    def send(self, x, y):
        if not isinstance(y, list):
            y = [y]

        if len(y) != len(self.traces):
            raise Exception(
                'You tried to set more y values (%d items) then traces available in channel %s (%d traces).' % (
                    len(y), self.name, len(self.traces)))

        line = json.dumps([x] + y)[1:-1]
        self.stream.write(line + "\n")
        self.job_backend.git.store_file('aetros/job/channel/' + self.name + '/last.csv', line)

        if self.kpi:
            self.job_backend.git.store_file('aetros/job/kpi/last.json', json.dumps(y[self.kpiTrace]))


class JobBackend:
    """
    :type event_listener: EventListener
    :type id: str: model name
    :type job_id: str: job id
    :type client: Client
    :type job: dict
    """

    def __init__(self, model_name=None, logger=None):
        self.event_listener = EventListener()

        self.log_file_handle = None
        self.config = {}
        self.job = {'parameters': {}}

        self.git = None

        self.ssh_stream = None
        self.model_name = model_name
        self.logger = logger

        self.client = None
        self.stream_log = None

        self.last_batch_time = time.time()
        self.start_time = time.time()
        self.current_epoch = 0
        self.current_batch = 0
        self.total_epochs = 0
        self.made_batches = 0
        self.batches_per_second = 0

        # indicates whether early_stop() has been called. Is called by reaching maxTimes or maxEpochs limitation.
        # This flag stops exiting with > 0, since the reach of a limitation is a valid exit.
        self.in_early_stop = False

        # done means: done, abort or crash method has been called.
        self.ended = False

        # when stop(wait_for_client=True) is called, we sync last messages.
        # this flag indicates that we are not yet ended completely but in the progress
        self.stopping = False
        # running means: the syncer client is running.
        self.running = False
        self.monitoring_thread = None

        if not self.logger:
            self.logger = logging.getLogger('aetros-job')
            coloredlogs.install(level=self.log_level, logger=self.logger)

        self.general_logger_stdout = GeneralLogger(job_backend=self)
        self.general_logger_error = GeneralLogger(job_backend=self, error=True)

        self.last_progress_call = None
        self.job_ids = []
        self.in_request = False
        self.stop_requested = False
        self.stop_requested_force = False

        self.kpi_channel = None
        self.registered = None
        self.progresses = {}

        self.event_listener.on('stop', self.external_stop)
        self.event_listener.on('aborted', self.external_aborted)
        self.event_listener.on('registration', self.on_registration)
        self.event_listener.on('registration_failed', self.on_registration_failed)
        self.event_listener.on('offline', self.on_client_offline)

        signal.signal(signal.SIGINT, self.on_signint)
        signal.signal(signal.SIGTERM, self.on_signint)
        signal.signal(signal.SIGUSR1, self.on_signusr1)

        self.pid = os.getpid()

        self.read_config(model_name)
        self.client = JobClient(self.config, self.event_listener, self.logger)
        self.git = Git(self.logger, self.client, self.host, self.config['storage_dir'], self.model_name)

    @property
    def log_level(self):
        if os.getenv('DEBUG') == '1':
            return 'DEBUG'

        return 'INFO'

    @property
    def host(self):
        return self.config['host']

    def on_registration_failed(self, params):
        self.logger.warning("Connecting to AETROS Trainer at %s failed. Reasons: %s" % (self.host, params['reason'],))
        if 'Permission denied' in params['reason']:
            self.logger.warning("Make sure you have saved your ssh pub key in your AETROS Trainer user account.")

    def on_client_offline(self, params):
        self.logger.warning("Could not establish a connection. We stopped automatic syncing.")
        self.logger.warning("You can publish later this job to AETROS Trainer using following command in this folder.")
        self.logger.warning("$ aetros push-job " + self.model_name + "/" + self.job_id)
        self.git.online = False

    def on_registration(self, params):
        if not self.registered:
            self.registered = True
            self.logger.info("Job %s/%s started." % (self.model_name, self.job_id))
            self.logger.info("Open http://%s/model/%s/job/%s to monitor the training." % (self.host, self.model_name, self.job_id))

            self.logger.debug('Git backend start')
            self.git.start()
        else:
            self.logger.info("Successfully reconnected.")

    def on_signusr1(self, signal, frame):
        self.logger.warning("USR1: backend (running=%s, ended=%s), client (online=%s, active=%s, registered=%s, "
                            "connected=%s, queue=%d), git (online=%s, active_thread=%s, last_push_time=%s)." % (
          str(self.running),
          str(self.ended),
          str(self.client.online),
          str(self.client.active),
          str(self.client.registered),
          str(self.client.connected),
          len(self.client.queue),
          str(self.git.online),
          str(self.git.active_thread),
          str(self.git.last_push_time),
        ))

    def on_signint(self, sig, frame):
        if self.stop_requested:
            self.stop_requested_force = True
            self.logger.warning('Force stopped')
            sys.exit(1)

        self.stop_requested = True

        self.logger.warning('Received SIGINT signal. Press CTRL+C again to force stop. Stopping ...')
        self.abort(True)
        sys.exit(1)

    def external_aborted(self, params):
        """
        Immediately abort the job
        """
        self.ended = True
        self.running = False

        if self.monitoring_thread:
            self.monitoring_thread.stop()

        self.git.stop()
        self.client.close()

        sys.exit(1)

    def external_stop(self, params):
        """
        Stop through GUI
        """
        if self.stop_requested:
            return

        self.logger.warning("Sent SIGINT through AETROS Trainer.")
        self.stop_requested_force = params
        os.kill(os.getpid(), signal.SIGINT)

    def setup_std_output_logging(self):
        """
        Overwrites sys.stdout and sys.stderr so we can send it additionally to Aetros.
        """
        sys.stdout = self.general_logger_stdout
        sys.stderr = self.general_logger_error

    def batch(self, batch, total, size=None):
        time_diff = time.time() - self.last_batch_time
        self.made_batches += 1

        if time_diff > 1 or batch == total:  # only each second second or last batch
            with self.git.batch_commit('BATCH'):
                self.set_system_info('batch', batch, True)
                self.set_system_info('batches', total, True)
                self.set_system_info('batchSize', size, True)

                self.batches_per_second = self.made_batches / time_diff
                self.made_batches = 0
                self.last_batch_time = time.time()

                if size:
                    self.set_system_info('samplesPerSecond', self.batches_per_second * size, True)

                epochs_per_second = self.batches_per_second / total  # all batches
                self.set_system_info('epochsPerSecond', epochs_per_second, True)

                elapsed = time.time() - self.start_time
                self.set_system_info('elapsed', elapsed, True)

                # self.git.store_file('aetros/job/times/elapsed.json', json.dumps(elapsed))

                if self.total_epochs:
                    eta = 0
                    if batch < total:
                        # time to end this epoch
                        if self.batches_per_second != 0:
                            eta = (total - batch) / self.batches_per_second

                    # time until all epochs are done
                    if self.total_epochs - (self.current_epoch - 1) > 0:
                        if epochs_per_second != 0:
                            eta += (self.total_epochs - (self.current_epoch - 1)) / epochs_per_second

                    self.git.store_file('aetros/job/times/eta.json', json.dumps(eta))

        self.current_batch = batch

    @property
    def job_settings(self):
        if 'settings' in self.job['config']:
            return self.job['config']['settings']
        return {}

    def create_progress(self, id, total_steps=100):
        if id in self.progresses:
            return self.progresses[id]

        class Controller():
            def __init__(self, git, id, total_steps=100):
                self.started = False
                self.stopped = False
                self.lock = Lock()
                self.id = id
                self.step = 0
                self.steps = total_steps
                self.eta = 0
                self.last_call = 0
                self.git = git
                self._label = id

                self.store()

            def store(self):
                info = {
                    'label': self._label,
                    'started': self.started,
                    'stopped': self.stopped,
                    'step': self.step,
                    'steps': self.steps,
                    'eta': self.eta,
                }
                self.git.store_file('aetros/job/progress/' + self.id + '.json', json.dumps(info))

            def label(self, label):
                self._label = label
                self.store()

            def start(self):
                if self.started is not False:
                    return

                self.started = time.time()
                self.last_call = time.time()
                self.store()

            def stop(self):
                if self.stopped is not False:
                    return

                self.stopped = time.time()
                self.store()

            def advance(self, steps=1):
                if steps <= 0:
                    return

                self.lock.acquire()
                if self.started is False:
                    self.start()

                took = (time.time() - self.last_call) / steps

                self.last_call = time.time()
                self.step += steps
                self.eta = took * (self.steps - self.step)

                if self.step >= self.steps:
                    self.stop()

                self.store()
                self.lock.release()

        self.progresses[id] = Controller(self.git, id, total_steps)

        return self.progresses[id]

    def progress(self, epoch, total):
        self.current_epoch = epoch
        self.total_epochs = total
        epoch_limit = False

        if 'maxEpochs' in self.job['config'] and self.job['config']['maxEpochs'] > 0:
            epoch_limit = True
            self.total_epochs = self.job['config']['maxEpochs']

        if epoch is not 0 and self.last_progress_call:
            # how long took it since the last call?
            time_per_epoch = time.time() - self.last_progress_call
            eta = time_per_epoch * (self.total_epochs - epoch)
            if epoch > total:
                eta = 0

            self.git.store_file('aetros/job/times/eta.json', json.dumps(eta))

            if time_per_epoch > 0:
                self.set_system_info('epochsPerSecond', 1 / time_per_epoch, True)

        self.set_system_info('epoch', epoch, True)
        self.set_system_info('epochs', self.total_epochs, True)
        self.last_progress_call = time.time()

        if epoch_limit and self.total_epochs > 0:
            if epoch >= self.total_epochs:
                self.logger.warning("\nMaxEpochs of %d/%d reached" % (epoch, self.total_epochs))
                self.early_stop()
                return

    def create_loss_channel(self, name, xaxis=None, yaxis=None, layout=None):
        """
        :param name: string
        :return: JobLossGraph
        """

        return JobLossChannel(self, name, xaxis, yaxis, layout)

    def create_channel(self, name, traces=None,
                       main=False, kpi=False, kpiTrace=0, max_optimization=True,
                       type=JobChannel.NUMBER,
                       xaxis=None, yaxis=None, layout=None):
        """
        :param name: str
        :param traces: None|list : per default create a trace based on "name".
        :param main: bool : whether this channel is visible in the job list as column for better comparison.

        :param kpi: bool : whether this channel is the KPI (key performance indicator).
                           Used for hyperparameter optimization. Only one channel can be a kpi. Only first trace used.
        :param kpiTrace: bool : if you have multiple traces, define which is the KPI. 0 based index.

        :param max_optimization: bool : whether the optimization maximizes or minmizes the kpi. Use max_optimization=False to
                                        tell the optimization algorithm that his channel minimizes a kpi, for instance the loss of a model.

        :param type: str : One of JobChannel.NUMBER, JobChannel.TEXT
        :param xaxis: dict
        :param yaxis: dict
        :param layout: dict
        """
        return JobChannel(self, name, traces, main, kpi, kpiTrace, max_optimization, type, xaxis, yaxis, layout)

    def start(self):
        if self.running:
            raise Exception('Job already running.')

        if not self.job_id:
            raise Exception('No job id found. Use create() first.')

        if not self.job:
            raise Exception('Job not loaded')

        on_shutdown.started_jobs.append(self)

        self.client.configure(self.model_name, self.job_id)
        if self.git.online:
            self.logger.debug('Job backend start')
            self.client.start()
        else:
            self.logger.debug('Job backend not started, since being online not detected.')

        self.git.commit_file('JOB', 'aetros/job/times/started.json', str(time.time()))
        self.job_add_status('progress', JOB_STATUS.PROGRESS_STATUS_STARTED)
        self.git.store_file('aetros/job/times/elapsed.json', str(0))

        self.stream_log = self.git.stream_file('aetros/job/log.txt')

        self.running = True
        self.ended = False
        self.collect_system_information()
        self.collect_environment()
        self.start_monitoring()

        self.detect_git_version()

    def detect_git_version(self):
        try:
            commit_sha = git.get_current_commit_hash()
            if commit_sha:
                self.set_system_info('git_version', commit_sha)

            current_branch = git.get_current_branch()
            if current_branch:
                self.set_system_info('git_branch', current_branch)
        except:
            pass

    def early_stop(self):
        """
        Stop when a limitatin is reached (like maxEpoch, maxtime)
        """
        self.in_early_stop = True
        self.done()
        os.kill(self.pid, signal.SIGINT)

    def start_monitoring(self):
        if not self.monitoring_thread:
            self.monitoring_thread = MonitoringThread(self)
            self.monitoring_thread.daemon = True
            self.monitoring_thread.start()

    def create_keras_callback(self, model,
                              insights=False, insights_x=None,
                              additional_insights_layer=[],
                              confusion_matrix=False, validation_data=None, validation_data_size=None):

        """

        :type validation_data: int|None: (x, y) or generator
        :type validation_data_size: int|None: Defines the size of validation_data, if validation_data is a generator
        """

        if insights and (insights_x is None or insights_x is False):
            raise Exception('Can not build Keras callback with active insights but with invalid `insights_x` as input.')

        if confusion_matrix and (validation_data is None or validation_data is False):
            raise Exception('Can not build Keras callback with active confusion_matrix but with invalid `validation_data` as input.')

        from aetros.Trainer import Trainer
        self.trainer = Trainer(self)

        self.trainer.model = model

        from aetros.KerasCallback import KerasCallback
        self.callback = KerasCallback(self, self.general_logger_stdout, force_insights=insights)
        self.callback.insights_x = insights_x
        self.callback.insight_layer = additional_insights_layer
        self.callback.confusion_matrix = confusion_matrix
        self.callback.set_validation_data(validation_data, validation_data_size)

        return self.callback

    def upload_keras_graph(self, model):
        from aetros.keras import model_to_graph
        import keras

        if keras.__version__[0] == '2':
            graph = model_to_graph(model)
            self.set_graph(graph)

    def on_shutdown(self):
        if self.stopping:
            return

        if self.stop_requested:
            if self.stop_requested_force:
                self.crash('Force stopped.')
            else:
                self.abort()
            return

        if not self.ended and hasattr(sys, 'last_value'):
            # sys.last_value contains a exception, when there was an uncaught one
            if isinstance(sys.last_value, KeyboardInterrupt):
                self.abort()
            else:
                self.crash(type(sys.last_value).__name__ + ': ' + str(sys.last_value))

        elif self.running:
            self.done()

        if self.in_early_stop:
            sys.exit(0)

    def done(self):
        if not self.running:
            return

        self.stop(JOB_STATUS.PROGRESS_STATUS_DONE, True)

    def stop(self, progress, wait_for_client=False):
        if self.monitoring_thread:
            self.monitoring_thread.stop()

        self.stopping = True
        self.logger.debug("stopping ...")
        self.ended = True
        self.running = False

        info_to_send_job = not wait_for_client
        if not info_to_send_job:
            info_to_send_job = len(self.client.queue) > 0 and not self.client.registered

        if self.git.online:
            if wait_for_client:
                self.logger.debug("client send last %d messages ..." % (len(self.client.queue),))
                self.client.wait_sending_last_messages()

        # store all store_file and stream_file as blob
        self.logger.debug("Git stopping ...")
        self.git.stop()

        if self.client.online:
            # server stores now all store_file and stream_file as blob as well,
            # wait for connection close to make sure server stored their files in server's git
            self.logger.debug("client stopping ...")
            self.client.end()

        # everything stopped. Server has all blobs, so we add the last progress
        # and do last git push.
        self.git.commit_file('STOP', 'aetros/job/status/progress.json', json.dumps(progress, default=invalid_json_values))

        # both sides should have same blobs, so we do now our final push
        self.logger.debug("git last push ...")
        successful_push = self.git.push()

        if not successful_push and info_to_send_job:
            self.logger.warning("Not all job information has been uploaded.")
            self.logger.warning("Please run following command to make sure your job is stored on the server.")
            self.logger.warning("$ aetros push-job " + self.model_name + "/" + self.job_id)

        # remove the index file
        self.git.clean_up()

        self.stopping = False
        self.logger.info("Stopped %s with last commit %s." % (self.git.ref_head, self.git.git_last_commit))

    def abort(self, wait_for_client_messages=True):
        if not self.running:
            return

        self.set_status('ABORTED')

        self.stop(JOB_STATUS.PROGRESS_STATUS_ABORTED, wait_for_client_messages)

    def crash(self, error=None):
        with self.git.batch_commit('CRASH'):
            self.set_status('CRASHED')
            self.git.commit_json_file('CRASH_REPORT_ERROR', 'aetros/job/crash/error', str(error) if error else '')
            self.git.commit_json_file('CRASH_REPORT_LAST_MESSAGE', 'aetros/job/crash/last_message', self.general_logger_error.last_messages)

        # we need to make sure that the server got the git crash before we report the crash progress
        self.git.push()

        self.job_add_status('progress', JOB_STATUS.PROGRESS_STATUS_CRASHED)

        self.logger.info('Crash report stored : ' + self.git.git_last_commit)
        self.stop(JOB_STATUS.PROGRESS_STATUS_CRASHED)

    def write_log(self, message):
        if self.stream_log and self.running:
            self.stream_log.write(message)

    def set_status(self, status):
        self.logger.info('Job status changed to %s ' % (status,))
        self.job_add_status('status', status)

    @property
    def job_id(self):
        return self.git.job_id

    def create(self, create_info=None, hyperparameter=None, server='local', insights=None):
        """
        Creates a new job in git and pushes it.

        :param create_info: from the api.create_job_info(id). Contains e.g. code, layer information for simple models
        :param hyperparameter: simple nested dict with key->value.
        :param server:
        :param insights: whether you want to activate insights
        """
        self.job['server'] = server
        self.job['config'] = copy.deepcopy(self.config)
        self.job['optimization'] = None
        self.job['type'] = 'python'

        if 'parameters' not in self.job['config']:
            self.job['config']['parameters'] = {}

        if create_info is not None:
            if 'config' not in create_info:
                print(create_info)
                raise Exception('Given create_info is not valid.')

            self.job['type'] = create_info['type']
            self.job['config'].update(create_info['config'])

            if 'simple' == self.job['type']:
                for k, v in six.iteritems(create_info['datasets']):
                    if 'python' == v['type']:
                        self.git.add_file('aetros/dataset/' + k + '.py', v['code'])
                        del v['code']

                    self.git.add_file('aetros/dataset/' + k + '.json', json.dumps(v))

                self.git.add_file('aetros/model.py', json.dumps(create_info['code']))
                self.git.add_file('aetros/layer.json', json.dumps(create_info['layers']))

        if insights is not None:
            self.job['config']['insights'] = insights

        if hyperparameter is not None:
            self.job['config']['parameters'].update(hyperparameter)

        if 'insights' not in self.job['config']:
            self.job['config']['insights'] = False

        self.git.create_job_id(self.job)
        self.job['id'] = self.job_id

        self.logger.info("Job created " + self.model_name + '/' + self.job_id + " with git ref " + self.git.ref_head)
        return self.job_id


    def is_simple_model(self):
        if not self.job:
            raise Exception('Job not loaded yet. Use load(id) first.')

        if 'type' in self.job:
            return self.job['type'] == 'simple'

        return False

    def read_config(self, model_name=None):
        self.config = read_config(logger=self.logger)

        self.logger.debug('config: ' + json.dumps(self.config))

        if model_name is None and 'model' not in self.config:
            raise Exception('No AETROS Trainer model name given. Specify it in aetros.backend.start_job("model/name") or in .aetros.yml `model: model/name`.')

        if 'models' in self.config:
            if model_name in self.config['models']:
                self.logger.debug('Merged config values for ' + model_name)
                self.config.update(self.config['models'][model_name])

            del self.config['models']

        # todo, read parameters from script command arguments

        if not self.model_name and ('model' in self.job or not self.job['model']):
            raise Exception('No model name given. Specify in .aetros.yml or in aetros.backend.start_job("model/name")')

        ssh_command = self.config['ssh']
        ssh_command += ' -o StrictHostKeyChecking=no'

        if self.config['ssh_key']:
            ssh_command += ' -i "' + os.path.expanduser(self.config['ssh_key']) + '"'

        import tempfile
        f = tempfile.NamedTemporaryFile(delete=False)
        f.write(six.b(ssh_command + ' "$@"'))
        f.close()
        os.environ['GIT_SSH'] = f.name
        os.chmod(f.name, 0o700)

        import atexit
        def delelete_git_ssh_file():
            os.unlink(f.name)

        atexit.register(delelete_git_ssh_file)

        self.logger.debug('SSH_COMMAND:'+ssh_command)

    def get_parameter(self, path, default=None):
        """
        Reads hyper parameter from job configuration. If nothing found, fallback to user config in .aetros.yml
        :param path: str 
        :param default: *
        :return: *
        """
        params = self.job['config']['parameters']

        if path in params:
            return params[path]

        try:
            # try first parameters from job creator
            return self.read_dict_by_path(params, path)
        except:
            # if not found in parameters from the creator (Trainer usually) then try user config
            try:
                return self.read_dict_by_path(self.config['parameters'], path)
            except:
                if default is None:
                    raise

        # not found and default is not None
        return default

    def read_dict_by_path(self, dictionary, path):
        if path in dictionary:
            return dictionary[path]
        elif '.' not in path:
            raise Exception('Parameter ' + path + ' not found and no default value given.')

        if '.' in path:
            path = path.split('.')
            current = dictionary
            for item in path:
                if item not in current:
                    raise Exception('Parameter ' + path + ' not found and no default value given.')

                current = current[item]

            return current

    def load(self, job_id):
        """
        Loads job and sets as current.
        :param job_id: int
        """

        # normally git would create a job_id, but we have already one
        # so download the ref and check it out.
        self.git.fetch_job(job_id)

        if not os.path.exists(self.git.work_tree + '/aetros/job.json'):
            raise Exception('Could not load aetros/job.json from git repository. Make sure you have created the job correctly.')

        with open(self.git.work_tree + '/aetros/job.json') as f:
            self.job = json.loads(f.read())

        print(str(self.job))

        if not self.job:
            raise Exception('Could not parse aetros/job.json from git repository. Make sure you have created the job correctly.')

        self.job['id'] = job_id

        if os.path.exists(self.git.work_tree + '/aetros/job/status/progress.json'):
            with open(self.git.work_tree + '/aetros/job/status/progress.json', 'r') as f:
                progress = float(f.read().decode('utf-8'))
        else:
            progress = 0

        if progress >= 2:
            self.logger.error('You can not restart an existing job that was already running. You need to restart the job through AETROS Trainer.')
            self.logger.error('You can alternatively reset the git reference to the root commit and force push that ref to reset the job.')
            sys.exit(1)

        self.logger.debug(str(self.job))

    def get_job_model(self):
        """
        Returns a new JobModel instance with current loaded job data attached.
        :return: JobModel
        """
        if not self.job:
            raise Exception('Job not loaded yet. Use load(id) first.')

        from aetros.JobModel import JobModel

        return JobModel(self.job_id, self.job, self.config['storage_dir'])

    def sync_weights(self):
        self.job_add_status('status', 'SYNC WEIGHTS')
        print("Sync weights ...")
        try:
            self.upload_weights('latest.hdf5', self.get_job_model().get_weights_filepath_latest(), with_status=True)
        except:
            pass
        print("Weights synced.")

    def job_add_status(self, key, value):
        self.git.store_file('aetros/job/status/' + key + '.json', json.dumps(value, default=invalid_json_values))

    def set_info(self, key, value, commit_end_of_job=False):
        if commit_end_of_job:
            self.git.store_file('aetros/job/info/' + key + '.json', json.dumps(value, default=invalid_json_values))
        else:
            self.git.commit_json_file('INFO ' + key, 'aetros/job/info/' + key, value)

    def set_graph(self, graph):
        self.git.commit_json_file('GRAPH', 'aetros/job/graph', graph)

    def set_system_info(self, key, value, commit_end_of_job=False):
        if commit_end_of_job:
            self.git.store_file('aetros/job/system/' + key + '.json', json.dumps(value, default=invalid_json_values))
        else:
            self.git.commit_json_file('SYSTEM_INFO ' + key, 'aetros/job/system/' + key, value)

    def commit_file(self, path, title=None):
        path = os.path.expanduser(path).strip()

        path = os.path.relpath(path, os.getcwd())
        if path.startswith('..'):
            raise Exception('Can not commit files that are out of the current working directory.')

        if './' == path[0: 2]:
            path = path[2:]

        with self.git.batch_commit('FILE ' + (title or path)):
            if os.path.isdir(path):
                for file in os.listdir(path):
                    self.commit_file(path + '/' + file)
                return

            if os.path.getsize(path) > 10 * 1024 * 1024 * 1024:
                raise Exception('Can not upload file bigger than 10MB')

            with open(path, 'rb') as f:
                contents = f.read()

            self.git.commit_file('FILE ' + (title or path), path, contents)

    def job_add_insight(self, x, images, confusion_matrix):
        converted_images = []
        info = {}
        for image in images:
            if not isinstance(image, JobImage):
                raise Exception('job_add_insight only accepts JobImage instances in images argument')

            converted_images.append({
                'id': image.id,
                'image': self.pil_image_to_jpeg(image.image)
            })
            info[image.id] = {
                'file': image.id+'.jpg',
                'label': image.label
            }

        with self.git.batch_commit('INSIGHT ' + str(x)):

            for image in converted_images:
                self.git.commit_file('INSIGHT ' + str(image['id']), 'aetros/job/insight/'+str(x)+'/image/'+image['id']+'.jpg', image['image'])

            self.git.commit_json_file('INSIGHT CONFUSION_MATRIX', 'aetros/job/insight/'+str(x)+'/confusion_matrix', confusion_matrix)
            self.git.commit_json_file('INSIGHT INFO', 'aetros/job/insight/' + str(x) + '/info', info)

    def pil_image_to_jpeg(self, image):
        buffer = six.BytesIO()

        image.save(buffer, format="JPEG", optimize=True, quality=70)
        return buffer.getvalue()

    def collect_environment(self):
        import socket
        import os
        import pip
        import platform

        env = {}

        import aetros
        env['aetros_version'] = aetros.__version__
        env['python_version'] = platform.python_version()
        env['python_executable'] = sys.executable

        env['hostname'] = socket.gethostname()
        env['variables'] = dict(os.environ)

        if 'API_KEY' in env['variables']:
            del env['variables']['API_KEY']

        env['pip_packages'] = sorted([[i.key, i.version] for i in pip.get_installed_distributions()])
        self.set_system_info('environment', env)

    def collect_system_information(self):
        import psutil

        mem = psutil.virtual_memory()

        with self.git.batch_commit('JOB_SYSTEM_INFORMATION'):
            self.set_system_info('memory_total', mem.total)

            on_gpu = False

            import aetros.cuda_gpu
            props = aetros.cuda_gpu.get_device_properties(0)
            if props:
                self.set_system_info('cuda_available', True)
                self.set_system_info('cuda_device_number', props['device'])
                self.set_system_info('cuda_device_name', props['name'])
                self.set_system_info('cuda_device_max_memory', props['memory'])

            self.set_system_info('on_gpu', on_gpu)

            import cpuinfo
            cpu = cpuinfo.get_cpu_info()
            self.set_system_info('cpu_name', cpu['brand'])
            self.set_system_info('cpu', [cpu['hz_actual_raw'][0], cpu['count']])



# Copyright (c) 2014-2016, NVIDIA CORPORATION.  All rights reserved.
from __future__ import absolute_import
from __future__ import print_function
import math
from six.moves import range

# Find the best implementation available
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

import numpy as np
import PIL.Image
import scipy.misc

# Library defaults:
#   PIL.Image:
#       size -- (width, height)
#   np.array:
#       shape -- (height, width, channels)
#       range -- [0-255]
#       dtype -- uint8
#       channels -- RGB
#   caffe.datum:
#       datum.data type -- bytes (uint8)
#       datum.float_data type -- float32
#       when decoding images, channels are BGR
#   DIGITS:
#       image_dims -- (height, width, channels)

# List of supported file extensions
# Use like "if filename.endswith(SUPPORTED_EXTENSIONS)"
SUPPORTED_EXTENSIONS = ('.png', '.jpg', '.jpeg', '.bmp', '.ppm')


def upscale(image, ratio):
    """
    return upscaled image array
    Arguments:
    image -- a (H,W,C) numpy.ndarray
    ratio -- scaling factor (>1)
    """
    if not isinstance(image, np.ndarray):
        raise ValueError('Expected ndarray')
    if ratio < 1:
        raise ValueError('Ratio must be greater than 1 (ratio=%f)' % ratio)
    width = int(math.floor(image.shape[1] * ratio))
    height = int(math.floor(image.shape[0] * ratio))
    channels = image.shape[2]
    out = np.ndarray((height, width, channels), dtype=np.uint8)
    for x, y in np.ndindex((width, height)):
        out[y, x] = image[int(math.floor(y / ratio)), int(math.floor(x / ratio))]
    return out


def resize_image(image, height, width,
                 channels=None,
                 resize_mode=None
                 ):
    """
    Resizes an image and returns it as a np.array
    Arguments:
    image -- a PIL.Image or numpy.ndarray
    height -- height of new image
    width -- width of new image
    Keyword Arguments:
    channels -- channels of new image (stays unchanged if not specified)
    resize_mode -- can be crop, squash, fill or half_crop
    """
    if resize_mode is None:
        resize_mode = 'squash'
    if resize_mode not in ['crop', 'squash', 'fill', 'half_crop']:
        raise ValueError('resize_mode "%s" not supported' % resize_mode)

    if channels not in [None, 1, 3]:
        raise ValueError('unsupported number of channels: %s' % channels)

    if isinstance(image, PIL.Image.Image):
        # Convert image mode (channels)
        if channels is None:
            image_mode = image.mode
            if image_mode == 'L':
                channels = 1
            elif image_mode == 'RGB':
                channels = 3
            else:
                raise ValueError('unknown image mode "%s"' % image_mode)
        elif channels == 1:
            # 8-bit pixels, black and white
            image_mode = 'L'
        elif channels == 3:
            # 3x8-bit pixels, true color
            image_mode = 'RGB'
        if image.mode != image_mode:
            image = image.convert(image_mode)
        image = np.array(image)
    elif isinstance(image, np.ndarray):
        if image.dtype != np.uint8:
            image = image.astype(np.uint8)
        if image.ndim == 3 and image.shape[2] == 1:
            image = image.reshape(image.shape[:2])
        if channels is None:
            if image.ndim == 2:
                channels = 1
            elif image.ndim == 3 and image.shape[2] == 3:
                channels = 3
            else:
                raise ValueError('invalid image shape: %s' % (image.shape,))
        elif channels == 1:
            if image.ndim != 2:
                if image.ndim == 3 and image.shape[2] == 3:
                    # color to grayscale
                    image = np.dot(image, [0.299, 0.587, 0.114]).astype(np.uint8)
                else:
                    raise ValueError('invalid image shape: %s' % (image.shape,))
        elif channels == 3:
            if image.ndim == 2:
                # grayscale to color
                image = np.repeat(image, 3).reshape(image.shape + (3,))
            elif image.shape[2] != 3:
                raise ValueError('invalid image shape: %s' % (image.shape,))
    else:
        raise ValueError('resize_image() expected a PIL.Image.Image or a numpy.ndarray')

    # No need to resize
    if image.shape[0] == height and image.shape[1] == width:
        return image

    # Resize
    interp = 'bilinear'

    width_ratio = float(image.shape[1]) / width
    height_ratio = float(image.shape[0]) / height
    if resize_mode == 'squash' or width_ratio == height_ratio:
        return scipy.misc.imresize(image, (height, width), interp=interp)
    elif resize_mode == 'crop':
        # resize to smallest of ratios (relatively larger image), keeping aspect ratio
        if width_ratio > height_ratio:
            resize_height = height
            resize_width = int(round(image.shape[1] / height_ratio))
        else:
            resize_width = width
            resize_height = int(round(image.shape[0] / width_ratio))
        image = scipy.misc.imresize(image, (resize_height, resize_width), interp=interp)

        # chop off ends of dimension that is still too long
        if width_ratio > height_ratio:
            start = int(round((resize_width - width) / 2.0))
            return image[:, start:start + width]
        else:
            start = int(round((resize_height - height) / 2.0))
            return image[start:start + height, :]
    else:
        if resize_mode == 'fill':
            # resize to biggest of ratios (relatively smaller image), keeping aspect ratio
            if width_ratio > height_ratio:
                resize_width = width
                resize_height = int(round(image.shape[0] / width_ratio))
                if (height - resize_height) % 2 == 1:
                    resize_height += 1
            else:
                resize_height = height
                resize_width = int(round(image.shape[1] / height_ratio))
                if (width - resize_width) % 2 == 1:
                    resize_width += 1
            image = scipy.misc.imresize(image, (resize_height, resize_width), interp=interp)
        elif resize_mode == 'half_crop':
            # resize to average ratio keeping aspect ratio
            new_ratio = (width_ratio + height_ratio) / 2.0
            resize_width = int(round(image.shape[1] / new_ratio))
            resize_height = int(round(image.shape[0] / new_ratio))
            if width_ratio > height_ratio and (height - resize_height) % 2 == 1:
                resize_height += 1
            elif width_ratio < height_ratio and (width - resize_width) % 2 == 1:
                resize_width += 1
            image = scipy.misc.imresize(image, (resize_height, resize_width), interp=interp)
            # chop off ends of dimension that is still too long
            if width_ratio > height_ratio:
                start = int(round((resize_width - width) / 2.0))
                image = image[:, start:start + width]
            else:
                start = int(round((resize_height - height) / 2.0))
                image = image[start:start + height, :]
        else:
            raise Exception('unrecognized resize_mode "%s"' % resize_mode)

        # fill ends of dimension that is too short with random noise
        if width_ratio > height_ratio:
            padding = (height - resize_height) / 2
            noise_size = (padding, width)
            if channels > 1:
                noise_size += (channels,)
            noise = np.random.randint(0, 255, noise_size).astype('uint8')
            image = np.concatenate((noise, image, noise), axis=0)
        else:
            padding = (width - resize_width) / 2
            noise_size = (height, padding)
            if channels > 1:
                noise_size += (channels,)
            noise = np.random.randint(0, 255, noise_size).astype('uint8')
            image = np.concatenate((noise, image, noise), axis=1)

        return image


def embed_image_html(image):
    """
    Returns an image embedded in HTML base64 format
    (Based on Caffe's web_demo)
    Arguments:
    image -- a PIL.Image or np.ndarray
    """
    if image is None:
        return None
    elif isinstance(image, PIL.Image.Image):
        pass
    elif isinstance(image, np.ndarray):
        image = PIL.Image.fromarray(image)
    else:
        raise ValueError('image must be a PIL.Image or a np.ndarray')

    # Read format from the image
    fmt = image.format
    if not fmt:
        # default to JPEG
        fmt = 'jpeg'
    else:
        fmt = fmt.lower()

    string_buf = StringIO()
    image.save(string_buf, format=fmt)
    data = string_buf.getvalue().encode('base64').replace('\n', '')
    return 'data:image/%s;base64,%s' % (fmt, data)


def add_bboxes_to_image(image, bboxes, color='red', width=1):
    """
    Draw rectangles on the image for the bounding boxes
    Returns a PIL.Image
    Arguments:
    image -- input image
    bboxes -- bounding boxes in the [((l, t), (r, b)), ...] format
    Keyword arguments:
    color -- color to draw the rectangles
    width -- line width of the rectangles
    Example:
    image = Image.open(filename)
    add_bboxes_to_image(image, bboxes[filename], width=2, color='#FF7700')
    image.show()
    """
    def expanded_bbox(bbox, n):
        """
        Grow the bounding box by n pixels
        """
        l = min(bbox[0][0], bbox[1][0])
        r = max(bbox[0][0], bbox[1][0])
        t = min(bbox[0][1], bbox[1][1])
        b = max(bbox[0][1], bbox[1][1])
        return ((l - n, t - n), (r + n, b + n))

    from PIL import Image, ImageDraw
    draw = ImageDraw.Draw(image)
    for bbox in bboxes:
        for n in range(width):
            draw.rectangle(expanded_bbox(bbox, n), outline=color)

    return image


def get_layer_vis_square(data,
                         allow_heatmap=True,
                         normalize=True,
                         min_img_dim=100,
                         max_width=1200,
                         channel_order='RGB',
                         colormap='jet',
                         ):
    """
    Returns a vis_square for the given layer data
    Arguments:
    data -- a np.ndarray
    Keyword arguments:
    allow_heatmap -- if True, convert single channel images to heatmaps
    normalize -- whether to normalize the data when visualizing
    max_width -- maximum width for the vis_square
    """
    if channel_order not in ['RGB', 'BGR']:
        raise ValueError('Unsupported channel_order %s' % channel_order)
    if data.ndim == 1:
        # interpret as 1x1 grayscale images
        # (N, 1, 1)
        data = data[:, np.newaxis, np.newaxis]
    elif data.ndim == 2:
        # interpret as 1x1 grayscale images
        # (N, 1, 1)
        data = data.reshape((data.shape[0] * data.shape[1], 1, 1))
    elif data.ndim == 3:
        if data.shape[0] == 3:
            # interpret as a color image
            # (1, H, W,3)
            if channel_order == 'BGR':
                data = data[[2, 1, 0], ...]  # BGR to RGB (see issue #59)
            data = data.transpose(1, 2, 0)
            data = data[np.newaxis, ...]
        else:
            # interpret as grayscale images
            # (N, H, W)
            pass
    elif data.ndim == 4:
        if data.shape[0] == 3:
            # interpret as HxW color images
            # (N, H, W, 3)
            data = data.transpose(1, 2, 3, 0)
            if channel_order == 'BGR':
                data = data[:, :, :, [2, 1, 0]]  # BGR to RGB (see issue #59)
        elif data.shape[1] == 3:
            # interpret as HxW color images
            # (N, H, W, 3)
            data = data.transpose(0, 2, 3, 1)
            if channel_order == 'BGR':
                data = data[:, :, :, [2, 1, 0]]  # BGR to RGB (see issue #59)
        else:
            # interpret as HxW grayscale images
            # (N, H, W)
            data = data.reshape((data.shape[0] * data.shape[1], data.shape[2], data.shape[3]))
    else:
        raise RuntimeError('unrecognized data shape: %s' % (data.shape,))

    return get_layer_vis_square_raw(data,
                         allow_heatmap,
                         normalize,
                         min_img_dim,
                         max_width,
                         colormap,
                         )

def get_image_tales(images, colormap='jet', min_img_dim=100, max_width=1000):

    padsize = 1
    # convert to float since we're going to do some math
    images = images.astype('float32')

    images -= images.min()
    if images.max() > 0:
        images /= images.max()
        images *= 255

    if images.ndim == 3:
        # they're grayscale - convert to a colormap
        redmap, greenmap, bluemap = get_color_map(colormap)

        red = np.interp(images * (len(redmap) - 1) / 255.0, range(len(redmap)), redmap)
        green = np.interp(images * (len(greenmap) - 1) / 255.0, range(len(greenmap)), greenmap)
        blue = np.interp(images * (len(bluemap) - 1) / 255.0, range(len(bluemap)), bluemap)

        # Slap the channels back together
        images = np.concatenate(
            (red[..., np.newaxis], green[..., np.newaxis], blue[..., np.newaxis]), axis=3)
        images = np.minimum(images, 255)
        images = np.maximum(images, 0)

    # convert back to uint8
    images = images.astype('uint8')

    # Compute the output image matrix dimensions
    n = int(np.ceil(np.sqrt(images.shape[0])))
    ny = n
    nx = n
    length = images.shape[0]
    if n * (n - 1) >= length:
        nx = n - 1

    # Add padding between the images
    padding = ((0, nx * ny - length), (0, padsize), (0, padsize)) + ((0, 0),) * (images.ndim - 3)
    padded = np.pad(images, padding, mode='constant', constant_values=0)

    # Tile the images beside each other
    tiles = padded.reshape(
        (ny, nx) + padded.shape[1:]).transpose((0, 2, 1, 3) + tuple(range(4, padded.ndim + 1)))
    tiles = tiles.reshape((ny * tiles.shape[1], nx * tiles.shape[3]) + tiles.shape[4:])

    return tiles

def get_layer_vis_square_raw(data,
                         allow_heatmap=True,
                         normalize=True,
                         min_img_dim=100,
                         max_width=1200,
                         colormap='jet',
                         ):
    # chop off data so that it will fit within max_width
    padsize = 0
    width = data.shape[2]
    if width > max_width:
        data = data[:1, :max_width, :max_width]
    else:
        if width > 1:
            padsize = 1
            width += 1
        n = max(max_width // width, 1)
        n *= n
        data = data[:n]

    if not allow_heatmap and data.ndim == 3:
        data = data[..., np.newaxis]

    vis = vis_square(data,
                     padsize=padsize,
                     normalize=normalize,
                     colormap=colormap
                     )

    # find minimum dimension and upscale if necessary
    _min = sorted(vis.shape[:2])[0]
    if _min < min_img_dim:
        # upscale image
        ratio = min_img_dim / float(_min)
        vis = upscale(vis, ratio)
    return vis


def vis_square(images,
               padsize=1,
               normalize=False,
               colormap='jet',
               ):
    """
    Visualize each image in a grid of size approx sqrt(n) by sqrt(n)
    Returns a np.array image
    (Based on Caffe's filter_visualization notebook)
    Arguments:
    images -- an array of shape (N, H, W) or (N, H, W, C)
            if C is not set, a heatmap is computed for the result
    Keyword arguments:
    padsize -- how many pixels go inbetween the tiles
    normalize -- if true, scales (min, max) across all images out to (0, 1)
    colormap -- a string representing one of the supported colormaps
    """
    assert 3 <= images.ndim <= 4, 'images.ndim must be 3 or 4'
    # convert to float since we're going to do some math
    images = images.astype('float32')
    if normalize:
        images -= images.min()
        if images.max() > 0:
            images /= images.max()
            images *= 255

    if images.ndim == 3:
        # they're grayscale - convert to a colormap
        redmap, greenmap, bluemap = get_color_map(colormap)

        red = np.interp(images * (len(redmap) - 1) / 255.0, range(len(redmap)), redmap)
        green = np.interp(images * (len(greenmap) - 1) / 255.0, range(len(greenmap)), greenmap)
        blue = np.interp(images * (len(bluemap) - 1) / 255.0, range(len(bluemap)), bluemap)

        # Slap the channels back together
        images = np.concatenate(
            (red[..., np.newaxis], green[..., np.newaxis], blue[..., np.newaxis]), axis=3)
        images = np.minimum(images, 255)
        images = np.maximum(images, 0)

    # convert back to uint8
    images = images.astype('uint8')

    # Compute the output image matrix dimensions
    n = int(np.ceil(np.sqrt(images.shape[0])))
    ny = n
    nx = n
    length = images.shape[0]
    if n * (n - 1) >= length:
        nx = n - 1

    # Add padding between the images
    padding = ((0, nx * ny - length), (0, padsize), (0, padsize)) + ((0, 0),) * (images.ndim - 3)
    padded = np.pad(images, padding, mode='constant', constant_values=255)

    # Tile the images beside each other
    tiles = padded.reshape(
        (ny, nx) + padded.shape[1:]).transpose((0, 2, 1, 3) + tuple(range(4, padded.ndim + 1)))
    tiles = tiles.reshape((ny * tiles.shape[1], nx * tiles.shape[3]) + tiles.shape[4:])

    if tiles.shape[-1] == 1:
        # grayscale to color
        tiles = np.dstack([tiles.squeeze()] * 3)

    return tiles


def get_color_map(name):
    """
    Return a colormap as (redmap, greenmap, bluemap)
    Arguments:
    name -- the name of the colormap. If unrecognized, will default to 'jet'.
    """
    redmap = [0]
    greenmap = [0]
    bluemap = [0]
    if name == 'white':
        # essentially a noop
        redmap = [0, 1]
        greenmap = [0, 1]
        bluemap = [0, 1]
    elif name == 'simple':
        redmap = [0, 1, 1, 1]
        greenmap = [0, 0, 1, 1]
        bluemap = [0, 0, 0, 1]
    elif name == 'hot':
        redmap = [0, 0.03968253968253968, 0.07936507936507936, 0.119047619047619, 0.1587301587301587, 0.1984126984126984, 0.2380952380952381, 0.2777777777777778, 0.3174603174603174, 0.3571428571428571, 0.3968253968253968, 0.4365079365079365, 0.4761904761904762, 0.5158730158730158, 0.5555555555555556, 0.5952380952380952,
                  0.6349206349206349, 0.6746031746031745, 0.7142857142857142, 0.753968253968254, 0.7936507936507936, 0.8333333333333333, 0.873015873015873, 0.9126984126984127, 0.9523809523809523, 0.992063492063492, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
        greenmap = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.03174603174603163, 0.0714285714285714, 0.1111111111111112, 0.1507936507936507, 0.1904761904761905, 0.23015873015873, 0.2698412698412698, 0.3095238095238093, 0.3492063492063491, 0.3888888888888888, 0.4285714285714284,
                    0.4682539682539679, 0.5079365079365079, 0.5476190476190477, 0.5873015873015872, 0.6269841269841268, 0.6666666666666665, 0.7063492063492065, 0.746031746031746, 0.7857142857142856, 0.8253968253968254, 0.8650793650793651, 0.9047619047619047, 0.9444444444444442, 0.984126984126984, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
        bluemap = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.04761904761904745, 0.1269841269841265,
                   0.2063492063492056, 0.2857142857142856, 0.3650793650793656, 0.4444444444444446, 0.5238095238095237, 0.6031746031746028, 0.6825396825396828, 0.7619047619047619, 0.8412698412698409, 0.92063492063492, 1]
    elif name == 'rainbow':
        redmap = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0.9365079365079367, 0.8571428571428572, 0.7777777777777777, 0.6984126984126986, 0.6190476190476191, 0.53968253968254, 0.4603174603174605, 0.3809523809523814, 0.3015873015873018, 0.2222222222222223, 0.1428571428571432,
                  0.06349206349206415, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.03174603174603208, 0.08465608465608465, 0.1375661375661377, 0.1904761904761907, 0.2433862433862437, 0.2962962962962963, 0.3492063492063493, 0.4021164021164023, 0.4550264550264553, 0.5079365079365079, 0.5608465608465609, 0.6137566137566139, 0.666666666666667]
        greenmap = [0, 0.03968253968253968, 0.07936507936507936, 0.119047619047619, 0.1587301587301587, 0.1984126984126984, 0.2380952380952381, 0.2777777777777778, 0.3174603174603174, 0.3571428571428571, 0.3968253968253968, 0.4365079365079365, 0.4761904761904762, 0.5158730158730158, 0.5555555555555556, 0.5952380952380952, 0.6349206349206349, 0.6746031746031745, 0.7142857142857142, 0.753968253968254, 0.7936507936507936,
                    0.8333333333333333, 0.873015873015873, 0.9126984126984127, 0.9523809523809523, 0.992063492063492, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0.9841269841269842, 0.9047619047619047, 0.8253968253968256, 0.7460317460317465, 0.666666666666667, 0.587301587301587, 0.5079365079365079, 0.4285714285714288, 0.3492063492063493, 0.2698412698412698, 0.1904761904761907, 0.1111111111111116, 0.03174603174603208, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        bluemap = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.01587301587301582, 0.09523809523809534, 0.1746031746031744, 0.2539682539682535,
                   0.333333333333333, 0.412698412698413, 0.4920634920634921, 0.5714285714285712, 0.6507936507936507, 0.7301587301587302, 0.8095238095238093, 0.8888888888888884, 0.9682539682539679, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    elif name == 'winter':
        greenmap = [0, 1]
        bluemap = [1, 0.5]
    else:
        if name != 'jet':
            print('Warning: colormap "%s" not supported. Using jet instead.' % name)
        redmap = [0, 0, 0, 0, 0.5, 1, 1, 1, 0.5]
        greenmap = [0, 0, 0.5, 1, 1, 1, 0.5, 0, 0]
        bluemap = [0.5, 1, 1, 1, 0.5, 0, 0, 0, 0]
    return 255.0 * np.array(redmap), 255.0 * np.array(greenmap), 255.0 * np.array(bluemap)



import platform
import ctypes


class CUDADeviceProperties(ctypes.Structure):
    # See $CUDA_HOME/include/cuda_runtime_api.h for the definition of
    # the cudaDeviceProp structypes.
    _fields_ = [
        ("name", ctypes.c_char * 256),
        ("totalGlobalMem", ctypes.c_size_t),
        ("sharedMemPerBlock", ctypes.c_size_t),
        ("regsPerBlock", ctypes.c_int),
        ("warpSize", ctypes.c_int),
        ("memPitch", ctypes.c_size_t),
        ("maxThreadsPerBlock", ctypes.c_int),
        ("maxThreadsDim", ctypes.c_int * 3),
        ("maxGridSize", ctypes.c_int * 3),
        ("clockRate", ctypes.c_int),
        ("totalConstMem", ctypes.c_size_t),
        ("major", ctypes.c_int),
        ("minor", ctypes.c_int),
        ("textureAlignment", ctypes.c_size_t),
        ("texturePitchAlignment", ctypes.c_size_t),
        ("deviceOverlap", ctypes.c_int),
        ("multiProcessorCount", ctypes.c_int),
        ("kernelExecTimeoutEnabled", ctypes.c_int),
        ("integrated", ctypes.c_int),
        ("canMapHostMemory", ctypes.c_int),
        ("computeMode", ctypes.c_int),
        ("maxTexture1D", ctypes.c_int),
        ("maxTexture1DMipmap", ctypes.c_int),
        ("maxTexture1DLinear", ctypes.c_int),
        ("maxTexture2D", ctypes.c_int * 2),
        ("maxTexture2DMipmap", ctypes.c_int * 2),
        ("maxTexture2DLinear", ctypes.c_int * 3),
        ("maxTexture2DGather", ctypes.c_int * 2),
        ("maxTexture3D", ctypes.c_int * 3),
        ("maxTexture3DAlt", ctypes.c_int * 3),
        ("maxTextureCubemap", ctypes.c_int),
        ("maxTexture1DLayered", ctypes.c_int * 2),
        ("maxTexture2DLayered", ctypes.c_int * 3),
        ("maxTextureCubemapLayered", ctypes.c_int * 2),
        ("maxSurface1D", ctypes.c_int),
        ("maxSurface2D", ctypes.c_int * 2),
        ("maxSurface3D", ctypes.c_int * 3),
        ("maxSurface1DLayered", ctypes.c_int * 2),
        ("maxSurface2DLayered", ctypes.c_int * 3),
        ("maxSurfaceCubemap", ctypes.c_int),
        ("maxSurfaceCubemapLayered", ctypes.c_int * 2),
        ("surfaceAlignment", ctypes.c_size_t),
        ("concurrentKernels", ctypes.c_int),
        ("ECCEnabled", ctypes.c_int),
        ("pciBusID", ctypes.c_int),
        ("pciDeviceID", ctypes.c_int),
        ("pciDomainID", ctypes.c_int),
        ("tccDriver", ctypes.c_int),
        ("asyncEngineCount", ctypes.c_int),
        ("unifiedAddressing", ctypes.c_int),
        ("memoryClockRate", ctypes.c_int),
        ("memoryBusWidth", ctypes.c_int),
        ("l2CacheSize", ctypes.c_int),
        ("maxThreadsPerMultiProcessor", ctypes.c_int),
        ("streamPrioritiesSupported", ctypes.c_int),
        ("globalL1CacheSupported", ctypes.c_int),
        ("localL1CacheSupported", ctypes.c_int),
        ("sharedMemPerMultiprocessor", ctypes.c_size_t),
        ("regsPerMultiprocessor", ctypes.c_int),
        ("managedMemSupported", ctypes.c_int),
        ("isMultiGpuBoard", ctypes.c_int),
        ("multiGpuBoardGroupID", ctypes.c_int),
        # Pad with extra space to avoid dereference crashes if future
        # versions of CUDA extend the size of this structypes.
        ("__future_buffer", ctypes.c_char * 4096)
    ]


def get_installed_devices():
    try:
        libcudart = get_libcudart()

        device_count = ctypes.c_int()
        libcudart.cudaGetDeviceCount(ctypes.byref(device_count))

        return device_count.value
    except:
        return 0


def get_memory(device):
    try:
        libcudart = get_libcudart()

        libcudart.cudaSetDevice(device)

        free = ctypes.c_size_t()
        total = ctypes.c_size_t()
        rc = libcudart.cudaMemGetInfo(ctypes.byref(free), ctypes.byref(total))
        if rc != 0:
            return None

        return free.value, total.value
    except:
        return None


def get_device_properties(device):
    try:
        libcudart = get_libcudart()

        properties = CUDADeviceProperties()
        rc = libcudart.cudaGetDeviceProperties(ctypes.byref(properties), device)
        if rc != 0:
            return None
    except:
        return None

    return {
        'device': device,
        'name': properties.name,
        'clock': properties.clockRate,
        'memory': properties.totalGlobalMem,
        'pciDeviceID': properties.pciDeviceID,
        'pciBusID': properties.pciBusID,
    }


def get_libcudart():
    system = platform.system()
    if system == "Linux":
        libcudart = ctypes.cdll.LoadLibrary("libcudart.so")
    elif system == "Darwin":
        libcudart = ctypes.cdll.LoadLibrary("libcudart.dylib")
    elif system == "Windows":
        libcudart = ctypes.windll.LoadLibrary("libcudart.dll")
    else:
        raise NotImplementedError("Cannot identify system.")

    version = ctypes.c_int()
    rc = libcudart.cudaRuntimeGetVersion(ctypes.byref(version))
    if rc != 0:
        raise ValueError("Could not get version")
    if version.value < 6050:
        raise NotImplementedError("CUDA version must be between >= 6.5")

    return libcudart



import subprocess

import six
from six.moves.urllib.parse import urlencode
from aetros.utils import read_config
import json

class ApiError(Exception):
    def __init__(self, message, reason):
        self.message = message
        self.reason = reason

    def __str__(self):
        return self.message + ', Reason: ' + self.reason

class ApiConnectionError(ApiError):
    pass

def request(path, query=None, body=None, config=None):
    query = query or {}

    if isinstance(query, dict):
        query = urlencode(query)

    if '?' in path:
        path += '&' + query
    else:
        path += '?' + query

    config = read_config() if config is None else config

    args = [config['ssh']] if isinstance(config['ssh'], six.string_types) else config['ssh']
    args += ['-o', 'StrictHostKeyChecking no']

    if config['ssh_key']:
        args += ['-i', config['ssh_key']]

    ssh_stream = subprocess.Popen(args + ['git@' + config['host'], 'api', path],
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE if body is not None else None, stdout=subprocess.PIPE)

    if body is not None:
        ssh_stream.stdin.write(json.dumps(body))
        ssh_stream.stdin.close()

    stderr = read(ssh_stream.stderr)
    stdout = read(ssh_stream.stdout)

    ssh_stream.wait()

    if ssh_stream.returncode != 0:
        if 'Connection ' in stderr:
            raise ApiConnectionError('Could not request api: ' + path, stderr)

        raise ApiError('Could not request api: ' + path, stderr)

    return stdout.decode('utf8')


def raise_response_exception(message, response):

    content = response.content
    error = 'unknown'
    error_message = ''

    if isinstance(content, six.string_types):
        try:
            content_parsed = json.loads(content)
            if 'error' in content_parsed:
                error = content_parsed['error']
            if 'message' in content_parsed:
                error_message = content_parsed['message']
        except:
            pass

    reason = 'StatusCode='+str(response.status_code)+', error: ' + str(error)+ ', message: ' + str(error_message)

    raise ApiError(message, reason)


def read(obj):
    r = six.b('')
    while True:
        buf = obj.read()
        if buf == six.b(''):
            break
        r += buf

    return r


def parse_json(content):
    a = json.loads(content)

    if 'error' in a:
        raise Exception('API request failed %s: %s.' % (a['error'], a['message']))

    return a


def model(model_name):
    content = request('model', {'id': model_name}).decode('utf-8')

    return parse_json(content)


def create_job_info(model_name, parameters=None, dataset_id=None):
    content = request('job/create-info', {'id': model_name}, {'parameters': parameters, 'datasetId': dataset_id}).decode('utf-8')

    return parse_json(content)



from __future__ import absolute_import
from __future__ import print_function
import argparse
import json
import os
import sys
import time
import psutil
import subprocess

import signal

import requests
import six
from requests.auth import HTTPBasicAuth

from aetros.api import raise_response_exception
from aetros.logger import GeneralLogger

from aetros.backend import EventListener, BackendClient
from aetros.utils import read_home_config


class ServerClient(BackendClient):
    def __init__(self, host, event_listener, logger):
        BackendClient.__init__(self, host, event_listener, logger)
        self.server_name = None

    def configure(self, server_name):
        self.server_name = server_name

    def on_connect(self, reconnect=False):
        self.send_message({'type': 'register_server', 'server': self.server_name})
        messages = self.wait_for_at_least_one_message()

        if not messages:
            return False

        message = messages.pop(0)
        if isinstance(message, dict) and 'a' in message:
            if "registration_failed" in message['a']:
                self.logger.error('Failed. ' + message['reason'])
                self.close()
                self.event_listener.fire('failed')
                return False

            if "already_registered" in message['a']:
                self.logger.error("Registration failed. This server is already registered.")
                return False

            if "registered" in message['a']:
                self.registered = True
                self.event_listener.fire('registration', {'username': message['username'], 'server': self.server_name})
                self.handle_messages(messages)
                return True

        self.logger.error("Registration of server %s failed due to protocol error." % (self.server_name,))
        if message:
            self.logger.error("Got server response: " + str(message))

        return False

    def handle_messages(self, messages):
        for message in messages:
            if not isinstance(message, dict):
                return

            if 'stop' in message:
                self.close()
                self.event_listener.fire('stop')

            if 'type' in message:
                if message['type'] == 'queue-jobs':
                    self.event_listener.fire('queue-jobs', message['jobs'])

                if message['type'] == 'unqueue-jobs':
                    self.event_listener.fire('unqueue-jobs', message['jobs'])

                if message['type'] == 'stop-job':
                    self.event_listener.fire('stop-job', message['id'])


class ServerCommand:
    model = None
    job_model = None

    def __init__(self, logger):
        self.logger = logger
        self.last_utilization = None
        self.last_net = {}
        self.nets = []
        self.server = None
        self.ending = False
        self.active = True
        self.queue = {}
        self.queuedMap = {}

        self.general_logger_stdout = None
        self.general_logger_stderr = None

        self.executed_jobs = 0
        self.max_parallel_jobs = 2
        self.max_jobs = 0
        self.ssh_key_path = None

        self.job_processes = []
        self.registered = False
        self.show_stdout = False

    def main(self, args):
        import aetros.const

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                         prog=aetros.const.__prog__ + ' server')
        parser.add_argument('name', nargs='?', help="Server name")
        parser.add_argument('--generate-ssh-key', help="Generates automatically a ssh key, register them in AETROS in your account, and delete them when the command exits.")
        parser.add_argument('--max-parallel', help="How many jobs should be run at the same time.")
        parser.add_argument('--max-jobs', help="How many jobs are allowed to run in total.")
        parser.add_argument('--host', help="Default trainer.aetros.com. Read from environment variable API_HOST.")
        parser.add_argument('--port', help="Default 8051. Read from environment variable API_PORT.")
        parser.add_argument('--show-stdout', action='store_true', help="Show all stdout of all jobs")

        parsed_args = parser.parse_args(args)

        if not parsed_args.name:
            parser.print_help()
            sys.exit()

        config = read_home_config()

        if parsed_args.max_parallel:
            self.max_parallel_jobs = int(parsed_args.max_parallel)
        if parsed_args.max_jobs:
            self.max_jobs = int(parsed_args.max_jobs)
        if parsed_args.show_stdout:
            self.show_stdout = True

        event_listener = EventListener()

        event_listener.on('registration', self.registration_complete)
        event_listener.on('failed', self.connection_failed)
        event_listener.on('queue-jobs', self.queue_jobs)
        event_listener.on('unqueue-jobs', self.unqueue_jobs)
        event_listener.on('queue-ok', self.queue_ok)
        event_listener.on('stop-job', self.stop_job)
        event_listener.on('close', self.on_client_close)

        signal.signal(signal.SIGUSR1, self.on_signusr1)

        ssh_key_registered = False
        if parsed_args.generate_ssh_key:

            self.ssh_key_path = os.path.expanduser('~/.ssh/id_' + parsed_args.name.replace('/', '__') + '_rsa')
            if not os.path.exists(self.ssh_key_path):
                self.logger.info('Generate SSH key')
                subprocess.check_output(['ssh-keygen', '-q', '-N', '', '-t', 'rsa', '-b', '4048', '-f', self.ssh_key_path])

            self.logger.info('Register SSH key at ' + config['host'])
            url = 'https://' + config['host'] + '/api/server/ssh-key'

            with open(self.ssh_key_path +'.pub', 'r') as f:
                data = {
                    'name': parsed_args.name,
                    'secure_key': parsed_args.generate_ssh_key,
                    'key': f.read(),
                }

                auth = None
                if 'auth_user' in config:
                    auth = HTTPBasicAuth(config['auth_user'], config['auth_pw'])

                response = requests.post(url, data, auth=auth, verify=config['ssl_verify'], headers={'Accept': 'application/json'})

                if response.status_code != 200:
                    raise_response_exception('Could not register SSH key in AETROS Trainer.', response)

                ssh_key_registered = response.content == 'true'

        def delete_ssh_key():
            with open(self.ssh_key_path +'.pub', 'r') as f:
                data = {
                    'secure_key': parsed_args.generate_ssh_key,
                    'key': f.read(),
                }
                self.logger.info('Delete SSH key at ' + config['host'])
                url = 'https://' + config['host'] + '/api/server/ssh-key/delete'

                auth = None
                if 'auth_user' in config:
                    auth = HTTPBasicAuth(config['auth_user'], config['auth_pw'])

                response = requests.post(url, data, auth=auth, verify=config['ssl_verify'], headers={'Accept': 'application/json'})

                if response.status_code != 200:
                    raise_response_exception('Could not delete SSH key in AETROS Trainer.', response)

                os.unlink(self.ssh_key_path)
                os.unlink(self.ssh_key_path +'.pub')

        if parsed_args.generate_ssh_key and ssh_key_registered:
            import atexit
            atexit.register(delete_ssh_key)

        if parsed_args.host:
            config['host'] = parsed_args.host

        if self.ssh_key_path:
            config['ssh_key'] = self.ssh_key_path

        self.server = ServerClient(config, event_listener, self.logger)

        self.general_logger_stdout = GeneralLogger(job_backend=self)
        self.general_logger_stderr = GeneralLogger(job_backend=self, error=True)

        sys.stdout = sys.__stdout__ = self.general_logger_stdout
        sys.stderr = sys.__stderr__ = self.general_logger_stderr

        self.server.configure(parsed_args.name)
        self.logger.info('Connecting to ' + config['host'])
        self.server.start()
        self.write_log("\n")

        try:
            while self.active:
                if self.registered:
                    self.server.send_message({'type': 'utilization', 'values': self.collect_system_utilization()})
                    self.process_queue()

                time.sleep(1)
        except KeyboardInterrupt:
            self.logger.warning('Aborted')
            self.stop()

    def on_signusr1(self, signal, frame):
        self.logger.info("%d queued, %d running, %d max" % (len(self.queue), len(self.job_processes), self.max_parallel_jobs))

    def on_client_close(self, params):
        self.active = False
        self.logger.warning('Closed')

    def write_log(self, message):
        self.server.send_message({'type': 'log', 'message': message})

    def stop(self):
        self.active = False

        for p in self.job_processes:
            p.kill()

        self.general_logger_stdout.flush()
        self.general_logger_stderr.flush()
        self.server.close()

    def end(self):
        self.ending = True

        for p in self.job_processes:
            p.wait()

        self.check_finished_jobs()
        self.stop()

    def connection_failed(self, params):
        self.active = False
        sys.exit(1)

    def stop_job(self, id):
        if id in self.queuedMap:
            job = self.queuedMap[id]
            self.logger.info("Queued job removed %s (priority: %s) " % (job['id'], job['priority']))

            # removing from the queue is enough, since the job process itself terminates it when job is aborted.
            if job in self.queue:
                self.queue[job['priority']].remove(job)

            del self.queuedMap[id]

    def unqueue_jobs(self, jobs):
        for id in jobs:
            if id in self.queuedMap:
                self.logger.info('Removed job %s from queue.' % (id, ))

                for priority in self.queue:
                    if self.queuedMap[id] in self.queue[priority]:
                        self.queue[priority].remove(self.queuedMap[id])

                del self.queuedMap[id]

    def queue_jobs(self, jobs):
        self.logger.debug('Got queue list with %d items.' % (len(jobs), ))

        for id in jobs.keys():
            self.check_finished_jobs()

            job = jobs[id]
            priority = job['priority']
            job['id'] = id

            if self.is_job_queued(id):
                self.logger.debug("Requested job %s is already known. Exclude that one." % (id, ))
                return

            self.logger.info("Queued job %s (priority:%d) in %s ..." % (
               job['id'], job['priority'],os.getcwd()
            ))

            self.queuedMap[job['id']] = job

            # add the job into the wait list
            if job['priority'] not in self.queue:
                self.queue[priority] = []

            self.queue[priority].append(job)

    def is_job_queued(self, id):
        return id in self.queuedMap

    def queued_count(self):
        i = 0
        for jobs in six.itervalues(self.queue):
            i += len(jobs)

        return i

    def is_job_running(self, id):
        for process in self.job_processes:
            job = getattr(process, 'job')
            if job['id'] == id:
                return True

        return False

    def queue_ok(self, id):
        """
        We queued the job, told the server so and server said we're ok to start the job now.

        :param id: 
        :return: 
        """
        job = self.queuedMap[id]
        priority = job['priority']

        self.logger.debug("Queued job confirmed %s (priority: %s) " % (job['id'], priority))

        # add the job into the wait list
        if job['priority'] not in self.queue:
            self.queue[priority] = []

        self.queue[priority].append(job)

    def check_finished_jobs(self):
        for process in self.job_processes:
            job = getattr(process, 'job')
            exit_code = process.poll()
            if exit_code is not None and exit_code > 0:
                reason = 'Failed job %s. Exit status: %s' % (job['id'], str(exit_code))
                self.logger.error(reason)
                self.server.send_message({'type': 'job-failed', 'id': job['id'], 'error': reason})
            elif exit_code is not None and exit_code == 0:
                self.logger.info('Finished job %s. Exit status: %s' % (job['id'], str(exit_code)))

            if exit_code is not None and job['id'] in self.queuedMap:
                del self.queuedMap[job['id']]

        # remove dead job processes
        self.job_processes = [x for x in self.job_processes if x.poll() is None]

    def process_queue(self):
        self.check_finished_jobs()

        if self.ending:
            return

        if len(self.job_processes) >= self.max_parallel_jobs:
            return

        if self.max_jobs and self.executed_jobs >= self.max_jobs:
            self.logger.warning('Limit of max jobs %d/%d reached. Waiting for active jobs to finish ...' % (self.executed_jobs, self.max_jobs))
            self.end()
            return

        # sort by priority: The higher the sooner the job starts
        for priority in sorted(self.queue, reverse=True):
            q = self.queue[priority]

            if len(q) > 0:
                # registered and free space for new jobs, so execute another one
                self.execute_job(q.pop(0))
                break

    def execute_job(self, job):
        self.logger.info("Execute job %s (priority=%s) in %s ..." % (job['id'], job['priority'], os.getcwd()))

        self.executed_jobs += 1

        with open(os.devnull, 'r+b', 0) as DEVNULL:
            my_env = os.environ.copy()

            if self.ssh_key_path is not None:
                my_env['AETROS_SSH_KEY'] = self.ssh_key_path

            args = [sys.executable, '-m', 'aetros', 'start', job['id']]
            self.logger.info('$ ' + ' '.join(args))
            self.server.send_message({'type': 'job-executed', 'id': job['id']})

            process = subprocess.Popen(args, bufsize=1,
                stdin=DEVNULL, stderr=subprocess.PIPE, stdout=subprocess.PIPE, env=my_env)

            self.general_logger_stdout.attach(process.stdout)
            self.general_logger_stderr.attach(process.stderr)

            setattr(process, 'job', job)
            self.job_processes.append(process)

    def registration_complete(self, params):
        self.registered = True

        # upon registration, we need to clear the queue, since the server sends us immediately
        # all to be enqueued jobs after registration/re-connection
        self.queue = {}
        self.queueMap = {}

        self.logger.info("As server %s under account %s registered." % (params['server'], params['username']))
        self.server.send_message({'type': 'system', 'values': self.collect_system_information()})

    def collect_system_information(self):
        values = {}
        mem = psutil.virtual_memory()
        values['memory_total'] = mem.total

        import cpuinfo
        cpu = cpuinfo.get_cpu_info()
        values['cpu_name'] = cpu['brand']
        values['cpu'] = [cpu['hz_actual_raw'][0], cpu['count']]
        values['nets'] = {}
        values['disks'] = {}
        values['boot_time'] = psutil.boot_time()

        for disk in psutil.disk_partitions():
            try:
                name = self.get_disk_name(disk[1])
                values['disks'][name] = psutil.disk_usage(disk[1]).total
            except:
                # suppress Operation not permitted
                pass

        try:
            for id, net in psutil.net_if_stats().items():
                if 0 != id.find('lo') and net.isup:
                    self.nets.append(id)
                    values['nets'][id] = net.speed or 1000
        except:
            # suppress Operation not permitted
            pass

        return values

    def get_disk_name(self, name):

        if 0 == name.find("/Volumes"):
            return os.path.basename(name)

        return name

    def collect_system_utilization(self):
        values = {}

        values['cpu'] = psutil.cpu_percent(interval=0.2, percpu=True)
        mem = psutil.virtual_memory()
        values['memory'] = mem.percent
        values['disks'] = {}
        values['jobs'] = {'parallel': self.max_parallel_jobs, 'enqueued': self.queued_count(), 'running': len(self.job_processes)}
        values['nets'] = {}
        values['processes'] = []

        for disk in psutil.disk_partitions():
            try:
                name = self.get_disk_name(disk[1])
                values['disks'][name] = psutil.disk_usage(disk[1]).used
            except:
                pass

        net_stats = psutil.net_io_counters(pernic=True)
        for id in self.nets:
            net = net_stats[id]
            values['nets'][id] = {
                'recv': net.bytes_recv,
                'sent': net.bytes_sent,
                'upload': 0,
                'download': 0
            }

            if id in self.last_net and self.last_utilization:
                values['nets'][id]['upload'] = (net.bytes_sent - self.last_net[id]['sent']) / (
                    time.time() - self.last_utilization)
                values['nets'][id]['download'] = (net.bytes_recv - self.last_net[id]['recv']) / (
                    time.time() - self.last_utilization)

            self.last_net[id] = dict(values['nets'][id])

        for p in psutil.process_iter():
            try:
                cpu = p.cpu_percent()
                if cpu > 1 or p.memory_percent() > 1:
                    values['processes'].append([
                        p.pid,
                        p.name(),
                        p.username(),
                        p.create_time(),
                        p.status(),
                        p.num_threads(),
                        p.memory_percent(),
                        cpu
                    ])
            except OSError:
                pass
            except psutil.Error:
                pass

        try:
            if hasattr(os, 'getloadavg'):
                values['loadavg'] = os.getloadavg()
            else:
                values['loadavg'] = ''
        except OSError:
            pass

        self.last_utilization = time.time()
        return values



from __future__ import division
from __future__ import absolute_import

import json
import os
import time
import psutil
from threading import Thread
import numpy as np
import signal


class MonitoringThread(Thread):
    def __init__(self, job_backend):
        Thread.__init__(self)

        self.job_backend = job_backend
        self.max_minutes = 0

        job = self.job_backend.job
        if 'maxTime' in job['config'] and job['config']['maxTime'] > 0:
            self.max_minutes = job['config']['maxTime']

        import sys
        if 'theano.sandbox' in sys.modules:
            # at this point, theano is already initialised, so we can use it to monitor the GPU.
            from theano.sandbox import cuda
            self.on_gpu = cuda.use.device_number is not None
        else:
            self.on_gpu = False

        self.stream = self.job_backend.git.stream_file('aetros/job/monitoring.csv')
        self.stream.write(json.dumps(["second", "cpu", "memory", "memory_gpu"])[1:-1] + "\n")
        self.second = 0
        self.started = time.time()
        self.running = True

    def stop(self):
        self.running = False

    def run(self):
        while self.running:
            self.monitor()
            time.sleep(0.05)

    def monitor(self):
        cpu_util = np.mean(psutil.cpu_percent(interval=1, percpu=True))
        mem = psutil.virtual_memory()

        minutes_run = (time.time() - self.started) / 60

        if not self.running:
            return

        if self.max_minutes > 0:
            if minutes_run > self.max_minutes:
                self.job_backend.logger.warning("\nMaxTime of %d/%d reached" % (minutes_run, self.max_minutes))
                self.job_backend.early_stop()

        gpu_memory_use = None

        import aetros.cuda_gpu
        info = aetros.cuda_gpu.get_memory(0)

        if info is not None:
            free, total = info
            gpu_memory_use = free/total*100

        self.stream.write(json.dumps([self.second, cpu_util, mem.percent, gpu_memory_use])[1:-1] + "\n")
        self.job_backend.git.store_file('aetros/job/times/elapsed.json', json.dumps(time.time() - self.started))

        self.second += 1
        pass


from __future__ import absolute_import
import argparse
import sys
import aetros.const
import os

class StartCommand:
    def __init__(self, logger):
        self.logger = logger

    def main(self, args):
        from aetros.starter import start
        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter, prog=aetros.const.__prog__ + ' start')
        parser.add_argument('name', nargs='?', help='the model name, e.g. aetros/mnist-network, or job id, e.g. user/modelname/0db75a64acb74c27bd72c22e359de7a4c44a20e5.')
        parser.add_argument('--insights', action='store_true', help="activates insights. Only for simple models.")
        parser.add_argument('--dataset', help="Dataset id when model has placeholders. Only for simple models with placeholders as input/output.")
        parser.add_argument('--gpu', action='store_true', help="Activates GPU if available. Only for Theano models.")
        parser.add_argument('--device', help="Which device index should be used. Default 0 (which means with --gpu => 'gpu0'). Only for Theano models.")
        parser.add_argument('--tf', action='store_true', help="Force TensorFlow as library. Only for simple models.")
        parser.add_argument('--th', action='store_true', help="Force Theano as library. Only for simple models.")
        parser.add_argument('--mp', help="Activates multithreading if available with given thread count. Only when Theano is active.")
        parser.add_argument('--no-hardware-monitoring', action='store_true', help="Deactivates hardware monitoring")
        parser.add_argument('--param', action='append', help="Sets a hyperparameter, example '--param name=value'. Multiple --param allowed.")

        parsed_args = parser.parse_args(args)

        if not parsed_args.name:
            parser.print_help()
            sys.exit()

        flags = os.environ['THEANO_FLAGS'] if 'THEANO_FLAGS' in os.environ else ''
        if parsed_args.gpu:
            if parsed_args.device:
                flags += ",device=gpu" + parsed_args.device
            else:
                flags += ",device=gpu"

        if parsed_args.mp:
            flags += ",openmp=True"
            os.environ['OMP_NUM_THREADS'] = parsed_args.mp

        os.environ['THEANO_FLAGS'] = flags

        if parsed_args.tf:
            os.environ['KERAS_BACKEND'] = 'tensorflow'

        if parsed_args.th:
            os.environ['KERAS_BACKEND'] = 'theano'

        hyperparameter = {}
        if parsed_args.param:
            for param in parsed_args.param:
                if '=' not in param:
                    raise Exception('--param ' + param+' does not contain a =. Please use "--param name=value"')

                name, value = param.split('=')
                hyperparameter[name] = value

        start(self.logger, parsed_args.name,
            hyperparameter=hyperparameter,
            dataset_id=parsed_args.dataset,
            insights=parsed_args.insights
        )



import json
import os
import subprocess

import signal
import six
from threading import Thread, Lock

import time

import sys

from aetros.api import ApiConnectionError
from aetros.utils import invalid_json_values


class GitCommandException(Exception):
    pass


class Git:
    """
    This class is used to store and sync all job data to local git or (if online) stream files directly to AETROS Trainer server.
    
    Git.stream_file and Git.store_file both stream new data directly to the server. At the end (Git.end) we commit
    the files locally and (if online) store the blob on the server's git. On a `git push` git detects that the content
    is already on our server which eliminates useless content transmissions.
    
    In either way (online or offline) all job information is stored in the local git repository and can be pushed any
    time to the AETROS git server. If the training happened in offline modus, one can push the job's ref (e.g. git push origin refs/aetros/job/<id>)
    to make the job available in AETROS Trainer later on.
    
    Job id is created always in the local git, except the job has been created through the AETROS Trainer interface.
    If created in AETROS Trainer, we retrieve the initial configuration of the job using `git pull origin refs/aetros/job/<id>` and read
    its `aetros/job.json` blob of the head tree.
    """
    def __init__(self, logger, client, git_host, storage_dir, model_name):
        self.logger = logger
        self.client = client
        self.git_host = git_host
        self.storage_dir = storage_dir
        self.model_name = model_name

        self.git_path = self.storage_dir + '/' + model_name + '.git'

        self.debug = False
        self.last_push_time = 0
        self.active_push = False
        self.index_path = None

        self.job_id = None
        self.online = True
        self.active_thread = False
        self.thread_push_instance = None

        self.git_batch_commit = False

        self.git_batch_commit_messages = []
        self.git_last_commit = None

        self.streamed_files = {}
        self.store_files = {}

        self.prepare_index_file()

        self.logger.debug("GIT_SSH='" + str(os.getenv('GIT_SSH'))+"'")
        self.git_name = None
        self.git_email = None

        import aetros.api
        try:
            user = json.loads(aetros.api.request('user-git'))

            self.git_name = user['name']
            self.git_email = user['email']
        except ApiConnectionError as e:
            self.online = False

        if not self.git_name:
            import getpass
            self.git_name = getpass.getuser()
            import socket
            self.git_email = self.git_name + '@' + socket.gethostname()

        # check if its a git repo
        if os.path.exists(self.git_path):
            out, code, err = self.command_exec(['remote'])
            if code != 0:
                raise Exception('Given git_path (%s) already exists and does not seem to be a git repository. Error: %s' % (self.git_path, err))
        else:
            os.makedirs(self.git_path)
            self.command_exec(['init'])
            self.command_exec(['remote', 'add', 'origin', self.git_url])

        # check if given repo_path is current folder.
        # check its origin remote and see if model_name matches
        origin_url = self.get_remote_url('origin')

        if origin_url and self.git_url not in origin_url:
            raise Exception("Given git_path (%s) points to a repository (%s) that is not the git repo of the model (%s). "
                            "You seem to switched between aetros.com and on-premise installation. "
                            "Please remove the Git repository." % (self.git_path, origin_url, self.git_url))

        if not os.path.exists(self.temp_path):
            os.makedirs(self.temp_path)

        self.logger.info("Started tracking of job files in git %s for remote %s" % (self.git_path, origin_url))

    def get_remote_url(self, origin_name):
        output = self.command_exec(['remote', '-v'], allowed_to_fail=True)[0].decode('utf-8').strip()

        import re
        match = re.match('^' + re.escape(origin_name) + '\t([^\s]+)', output)
        if match:
            return match.group(1)

    @property
    def work_tree(self):
        return self.storage_dir + '/' + self.model_name + '/' + self.job_id

    @property
    def env(self):
        my_env = os.environ.copy()
        if self.index_path:
            my_env['GIT_INDEX_FILE'] = self.index_path

        return my_env

    def thread_push(self):
        while self.active_thread:
            try:
                if self.job_id and self.online and self.active_push:
                    start = time.time()
                    self.command_exec(['push', '-f', 'origin', self.ref_head])
                    self.last_push_time = time.time() - start

                time.sleep(1)
            except SystemExit:
                return
            except KeyboardInterrupt:
                return

    @property
    def temp_path(self):
        return self.git_path + '/temp'

    @property
    def ref_head(self):
        return 'refs/aetros/job/' + self.job_id

    @property
    def git_url(self):
        return 'git@%s:%s.git' % (self.git_host, self.model_name)

    def command_exec(self, command, inputdata=None, allowed_to_fail=False):
        interrupted = False

        if isinstance(inputdata, six.string_types):
            inputdata = six.b(inputdata)

        if command[0] != 'git':
            base_command = ['git', '--bare', '--git-dir', self.git_path]
            base_command += ['-c', 'user.name=' + self.git_name]
            base_command += ['-c', 'user.email=' + self.git_email]
            command = base_command + command

        p = None
        stdoutdata = ''
        stderrdata = ''

        def preexec_fn():
            # Ignore the SIGINT signal by setting the handler to the standard
            # signal handler SIG_IGN.
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            signal.signal(signal.SIGINT, signal.SIG_IGN)

        try:
            p = subprocess.Popen(
                command, preexec_fn=preexec_fn, bufsize=0,
                stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=self.env
            )

            stdoutdata, stderrdata = p.communicate(inputdata)
        except KeyboardInterrupt:
            try:
                if p is not None:
                    stdoutdata = p.stdout.read()
                    stderrdata = p.stderr.read()
            except:
                pass
            interrupted = True

        input = ''

        try:
            input = inputdata.decode('utf-8')
        except:
            input = 'binary'

        try:
            stderrdata = stderrdata.decode('utf-8')
        except: pass

        message = "Git command: " + (' '.join(command))
        # message += "\nstdout: " + str(stdoutdata.decode('utf-8')) + ', stderr: '+ str(stderrdata)

        self.logger.debug(message)

        if 'Connection refused' in stderrdata or 'Permission denied' in stderrdata:
            if 'Permission denied' in stderrdata:
                self.logger.warning("You have no permission to push to that model.")

            self.online = False
            self.logger.error(stderrdata)
            return '', 1, ''

        if not interrupted and not allowed_to_fail and p is not None and p.returncode != 0:
            raise GitCommandException('Command failed: ' + ' '.join(command) + ', code: ' + str(p.returncode)+"\nstdout: '" + str(stdoutdata)+"',\nstderr: '" + str(stderrdata)+"', input="+str(inputdata))

        return stdoutdata, p.returncode, stderrdata

    def prepare_index_file(self):
        """
        Makes sure that GIT index file we use per job (by modifying environment variable GIT_INDEX_FILE)
        is not locked and empty. Git.fetch_job uses `git read-tree` to updates this index. For new jobs, we start
        with an empty index - that's why we delete it every time.
        """
        import tempfile
        h, path = tempfile.mkstemp('aetros-git')

        self.index_path = path

        # we give git a unique file path for that index. However, git expect it to be non-existent for empty indexes.
        # empty file would lead to "fatal: index file smaller than expected"
        os.close(h)
        os.unlink(self.index_path)

        self.logger.debug('GIT_INDEX_FILE created at ' + self.index_path)

    def fetch_job(self, job_id):
        """
        Fetch the current job reference (refs/aetros/job/<id>) from origin and read its tree to the current git index.
        
        :type job_id: str 
        """
        self.job_id = job_id

        self.logger.info("Git fetch job reference %s" % (self.ref_head, ))
        out, code, err = self.command_exec(['ls-remote', 'origin', self.ref_head])

        if code:
            self.logger.error('Could not find the job ' + job_id + ' on the server. Are you online and does the job exist?')
            sys.exit(1)

        try:
            self.command_exec(['fetch', '-f', '-n', 'origin', self.ref_head+':'+self.ref_head])
        except:
            self.logger.error("Could not load job information for " + job_id + '. You need to be online to start pre-configured jobs.')
            raise

        self.git_last_commit = job_id
        self.command_exec(['update-ref', self.ref_head, self.git_last_commit])
        self.command_exec(['read-tree', self.ref_head])

        # make sure we have checked out all files we have added until now. Important for simple models, so we have the
        # actual model.py and dataset scripts.
        if not os.path.exists(self.work_tree):
            os.makedirs(self.work_tree)

        self.logger.info('Working directory in ' + self.work_tree)
        self.command_exec(['--work-tree', self.work_tree, 'checkout', self.ref_head, '.'])

    def create_job_id(self, data):
        """
        Create a new job id and reference (refs/aetros/job/<id>) by creating a new commit with empty tree. That
        root commit is the actual job id. A reference is then created to the newest (head) commit of this commit history.
        The reference will always be updated once a new commit is added.
        """
        self.add_file('aetros/job.json', json.dumps(data))
        tree_id = self.write_tree()

        self.job_id = self.command_exec(['commit-tree', '-m', "JOB_CREATED", tree_id])[0].decode('utf-8').strip()
        self.git_last_commit = self.job_id

        out, code, err = self.command_exec(['show-ref', self.ref_head], allowed_to_fail=True)
        if not code:
            self.logger.warning("Generated job id already exists, because exact same experiment values given. Ref " + self.ref_head)

        self.command_exec(['update-ref', self.ref_head, self.git_last_commit])

        # make sure we have checkedout all files we have added until now. Important for simple models, so we have the
        # actual model.py and dataset scripts.
        if not os.path.exists(self.work_tree):
            os.makedirs(self.work_tree)

        self.command_exec(['--work-tree', self.work_tree, 'checkout', self.ref_head])

        return self.job_id

    def start(self):
        """
        Start the git push thread
        """
        if self.active_thread:
            return

        self.active_thread = True
        self.active_push = True

        self.thread_push_instance = Thread(target=self.thread_push)
        self.thread_push_instance.daemon = True
        self.thread_push_instance.start()

    def stop(self):
        """
        Stops the `git push` thread and commits all streamed files (Git.store_file and Git.stream_file), followed
        by a final git push.
        
        This removes also the current git index file. You can not start the process again.
        """
        self.active_thread = False

        with self.batch_commit('STREAM_END'):
            for path, handle in six.iteritems(self.streamed_files.copy()):
                full_path = self.temp_path + '/stream-blob/' + self.job_id + '/' + path
                handle.seek(0)
                self.commit_file(path, path, handle.read())
                handle.close()
                os.unlink(full_path)
            self.streamed_files = {}

        with self.batch_commit('STORE_END'):
            for path, bar in six.iteritems(self.store_files.copy()):
                full_path = self.temp_path + '/store-blob/' + self.job_id + '/' + path
                self.commit_file(path, path, open(full_path, 'r').read())
                os.unlink(full_path)
            self.store_files = {}

    def clean_up(self):
        if os.path.exists(self.index_path):
            os.remove(self.index_path)

    def batch_commit(self, message):
        """
        Instead of committing a lot of small commits you can batch it together using this controller.
        
        Example:
        
        with git.batch_commit('BATCHED'):
            git.commit_file('my commit 1', 'path/to/file', 'content from file')
            git.commit_json_file('[1, 2, 3]', 'path/to/file2', 'json array') 
            
        Withing the `with` block you can use group the method calls of `commit_file` and `commit_json_file`, and every other
        method calling this two methods.
        
        :type message: str 
        :return: with controller to be used with Python's `with git.batch_commit():`
        """
        class controlled_execution:
            def __init__(self, job, message):
                self.job = job
                self.message = message

            def __enter__(self):
                self.job.git_batch_commit = True

            def __exit__(self, type, value, traceback):
                self.job.git_batch_commit = False

                # if nothing committed, we return early
                if not self.job.git_batch_commit_messages: return

                commit_message = self.message
                if self.job.git_batch_commit_messages:
                    commit_message = commit_message + "\n\n" + "\n".join(self.job.git_batch_commit_messages)
                self.job.git_batch_commit_messages = []

                self.job.commit_index(commit_message)

        return controlled_execution(self, message)

    def get_empty_tree_id(self):
        """
        Returns the famous empty tree id. To be used in creating a new empty root commit without any files.

        :rtype: str 
        """
        return self.command_exec(['hash-object', '--stdin', '-ttree'], '')[0].decode('utf-8').strip()

    def store_file(self, path, data):
        """
        Store the file in temp folder and stream it to server if online. 
        
        This makes sure that we have all newest data of this file on the server directly. 
        
        This method always overwrites the content of path. If you want to append always the content, 
        use Git.stream_file() instead.
        
        At the end of the job, the content the server received is stored as git blob on the server. It is then committed 
        locally and pushed. Git detects that the server already has the version (through the continuous streaming)
        and won't push it again.
        
        :param path: 
        :param content: 
        :return: 
        """

        full_path = self.temp_path + '/store-blob/' + self.job_id + '/' + path
        if not os.path.exists(os.path.dirname(full_path)):
            os.makedirs(os.path.dirname(full_path))

        open(full_path, 'w+').write(data)
        self.store_files[path] = True

        if self.online:
            self.client.send({'type': 'store-blob', 'path': path, 'data': data})

    def stream_file(self, path):
        """
        Create a temp file, stream it to the server if online and append its content using the write() method. 
        This makes sure that we have all newest data of this file on the server directly.
        
        At the end of the job, the content the server received is stored as git blob on the server. It is then committed 
        locally and pushed. Git detects that the server already has the version (through the continuous streaming)
        and won't push it again. Very handy for rather large files that will append over time (like channel data, logs)
        
        Example:
        
        self.log_stream = git.stream_file('log.txt')
        
        self.log_stream.write("new line\n");
        self.log_stream.write("another line\n");
        
        :param path: 
        :rtype: Stream class
        :return Returns a instance with a `write(data)` method.
        """

        # create temp file
        # open temp file

        # register stream file and write locally
        # on end() git_commit that file

        # create socket connection to server
        # stream file to server
        # on end() send server end signal, so he can store its content in git as blob as well. A git push would detect
        # that both sides have the same content already, except when server connection broke between start() and end().

        # return handler to write to this file

        full_path = self.temp_path + '/stream-blob/' + self.job_id + '/' + path
        if not os.path.exists(os.path.dirname(full_path)):
            os.makedirs(os.path.dirname(full_path))

        handle = open(full_path, 'w+')
        self.streamed_files[path] = handle

        class Stream():
            def __init__(self, handle, git):
                self.handle = handle
                self.git = git

            def write(self, data):
                try:
                    handle.write(data)
                    if self.git.online:
                        self.git.client.send({'type': 'stream-blob', 'path': path, 'data': data})
                except:
                    pass

        return Stream(handle, self)

    def write_blob(self, content):
        return self.command_exec(['hash-object', '-w', "--stdin"], content)[0].decode('utf-8').strip()

    def add_index(self, mode, blob_id, path):
        """
        Add new entry to the current index
        :param tree: 
        :return: 
        """
        self.command_exec(['update-index', '--add', '--cacheinfo', mode, blob_id, path])

    def write_tree(self):
        """
        Writes the current index into a new tree
        :return: the tree sha
        """
        return self.command_exec(['write-tree'])[0].decode('utf-8').strip()

    def commit_json_file(self, message, path, content):
        return self.commit_file(message, path + '.json', json.dumps(content, default=invalid_json_values))

    def add_file(self, path, content):
        """
        Add a new file as blob in the storage and add its tree entry into the index.
        
        :param path: str
        :param content: str
        """
        blob_id = self.write_blob(content)
        self.add_index('100644', blob_id, path)

    def commit_file(self, message, path, content):
        """
        Add a new file as blob in the storage, add its tree entry into the index and commit the index.
         
        :param message: str 
        :param path: str
        :param content: str
        :return: 
        """
        self.add_file(path, content)

        if not self.git_batch_commit:
            self.commit_index(message)
        else:
            self.git_batch_commit_messages.append(message)

    def push(self):
        """
        Push all changes to origin
        """
        try:
            self.command_exec(['push', 'origin', '-f', self.ref_head])
            return True
        except:
            return False

    def commit_index(self, message):
        """
        Commit the current index.
        :param message: str
        :return: str the generated commit sha
        """
        tree_id = self.write_tree()

        args = ['commit-tree', tree_id]

        if self.git_last_commit:
            # if not define, it creates a new root commit
            args += ['-p', self.git_last_commit]

        self.git_last_commit = self.command_exec(args, message)[0].decode('utf-8').strip()

        # update ref
        self.command_exec(['update-ref', self.ref_head, self.git_last_commit])

        return self.git_last_commit

    def git_job_last_commit_sha(self):
        if not self.job_id:
            raise Exception('Can not fetch last commit sha, when no job_id is set.')

        out, code = self.command_exec(['show-ref', self.ref_head])
        if not code:
            return out.decode('utf-8').strip().split(' ')[0]
        else:
            raise Exception('Job ref not created yet. ' + self.ref_head)

    def git_read(self, path):
        return self.command_exec(['cat-file', '-p', self.ref_head+':'+path])



from __future__ import absolute_import
from __future__ import print_function
import argparse

import sys

from aetros import api

class ApiCommand:
    def __init__(self, logger):
        self.logger = logger
        self.client = None
        self.registered = False
        self.active = True

    def main(self, args):
        import aetros.const

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                description="You can provide json in stdin to issue a POST call", prog=aetros.const.__prog__ + ' run')
        parser.add_argument('path', nargs='?', help="Request path + query, e.g. model/settings?name=owner/name")
        parser.add_argument('--method', nargs='?', help="Per default GET, if stdin data is given a POST. Alternatively provide a HTTP verb.")

        parsed_args = parser.parse_args(args)

        if not parsed_args.path:
            parser.print_help()
            sys.exit(1)

        sys.stdout.write(api.request(parsed_args.path))


__version__ = '0.10.1'
__prog__ = "aetros"

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

class JOB_STATUS:
    PROGRESS_STATUS_CREATED = 0
    PROGRESS_STATUS_QUEUED = 1
    PROGRESS_STATUS_STARTED = 2
    PROGRESS_STATUS_DONE = 3
    PROGRESS_STATUS_ABORTED = 4
    PROGRESS_STATUS_CRASHED = 5


from __future__ import print_function, division

from __future__ import absolute_import
import io
import logging
import os
import pprint
import re
import shutil
import signal
import subprocess
import sys
import traceback

from aetros import api
from aetros.utils import git, read_home_config, unpack_full_job_id
from . import keras_model_utils
from .backend import JobBackend
from .Trainer import Trainer
from .keras_model_utils import ensure_dir
import six


class GitCommandException(Exception):
    cmd = None


def start(logger, full_id, hyperparameter=None, dataset_id=None, server='local', insights=False):
    """
    Starts the training process with all logging of a job_id
    :type id: string : job id or model name
    """
    id = None

    if full_id.count('/') == 1:
        owner, name = full_id.split('/')
    elif full_id.count('/') >= 2:
        owner, name, id = unpack_full_job_id(full_id)
    else:
        logger.error("Invalid id %s given. Supported formats: owner/modelName or owner/modelName/jobId." % (full_id, ))
        sys.exit(1)

    job_backend = JobBackend(model_name=owner + '/' + name)
    if id:
        job_backend.load(id)
    else:
        try:
            create_info = api.create_job_info(full_id, hyperparameter, dataset_id)
        except api.ApiError as e:
            if 'Connection refused' in e.reason:
                logger.error("You are offline")
            logger.error("Can not start new job without knowing what model type it is. "
                         "Use your script directly if its a Python model.")
            raise

        if not create_info or not create_info['id']:
            raise Exception('Could not fetch model information. Are you online and have access to the given model?')

        job_backend.create(create_info=create_info, hyperparameter=hyperparameter, server=server, insights=insights)

    if not len(job_backend.get_job_model().config):
        raise Exception('Job does not have a configuration. Make sure you created the job via AETROS Trainer.')

    if job_backend.is_simple_model():
        start_keras(logger, job_backend)
    else:
        start_custom(logger, job_backend)


def start_custom(logger, job_backend):
    job_model = job_backend.get_job_model()
    config = job_model.config

    custom_git = False

    if 'gitCustom' in config and config['gitCustom']:
        custom_git = config['gitCustom']

    if custom_git and 'sourceGitUrl' not in config or not config['sourceGitUrl']:
        raise Exception('Server git url is not configured. Aborted')

    if 'sourcePythonScript' not in config or not config['sourcePythonScript']:
        raise Exception('Server python script is not configured. Aborted')

    python_script = config['sourcePythonScript']
    git_tree = 'master'

    if custom_git:
        git_url = config['sourceGitUrl']
    else:
        user_config = read_home_config()
        git_url = 'git@' + user_config['host'] + ':' + job_backend.model_name + '.git'

    if 'sourceGitTree' in config and config['sourceGitTree']:
        git_tree = config['sourceGitTree']

    work_tree = job_backend.git.work_tree

    my_env = os.environ.copy()
    if 'PYTHONPATH' not in my_env:
        my_env['PYTHONPATH'] = ''
    my_env['PYTHONPATH'] += ':' + os.getcwd()
    my_env['AETROS_MODEL_NAME'] = job_backend.model_name
    my_env['AETROS_JOB_ID'] = job_backend.job_id

    logger.info("Setting up git repository %s in %s" % (git_url, work_tree))
    logger.info("Using git tree of '%s'" % (git_tree, ))

    try:
        if os.path.exists(work_tree):
            shutil.rmtree(work_tree)

        args = ['git', 'clone', git_url, work_tree]
        code = subprocess.call(args, stderr=sys.stderr, stdout=sys.stdout)
        if code != 0:
            raise Exception('Could not clone repository %s to %s' %(git_url, work_tree))

        # make sure the requested branch is existent in local git. Target FETCH_HEAD to this branch.
        git_execute(logger, work_tree, ['fetch', 'origin', git_tree])
        git_execute(logger, work_tree, ['checkout', git_tree])

    except GitCommandException as e:
        raise Exception('Could not run "%s" for repository %s in %s. Look at previous output.' %(e.cmd, git_url, work_tree))

    args = (sys.executable, python_script)
    logger.info("Model source code checked out.")
    logger.info("-----------")
    logger.info("-----------")
    logger.info("Switch working directory to " + work_tree)
    logger.warning("$ %s %s" % args)

    try:
        subprocess.Popen(args, close_fds=True, env=my_env, cwd=work_tree).wait()
    except KeyboardInterrupt:
        logger.warning("Job aborted.")
        sys.exit(1)


def git_execute(logger, repo_path, args):
    args = ['git', '--git-dir', repo_path + '/.git', '--work-tree', repo_path] + args
    logger.info("$ %s" % (' '.join(args), ))

    p = subprocess.Popen(args, stderr=sys.stderr, stdout=sys.stdout)
    p.wait()

    if p.returncode != 0:
        exception = GitCommandException("Git command returned not 0. " + (' '.join(args)))
        exception.cmd = (' '.join(args))
        raise exception


def start_keras(logger, job_backend):
    # we need to import keras here, so we know which backend is used (and whether GPU is used)
    os.chdir(job_backend.git.work_tree)

    job_backend.setup_std_output_logging()
    job_backend.start()

    # all our shapes are Tensorflow schema. (height, width, channels)
    from keras import backend as K
    if hasattr(K, 'set_image_dim_ordering'):
        K.set_image_dim_ordering('tf')

    job_model = job_backend.get_job_model()

    from .KerasCallback import KerasCallback
    trainer = Trainer(job_backend)
    keras_logger = KerasCallback(job_backend, job_backend.general_logger_stdout)

    try:
        logger.info("Setup simple job")
        keras_model_utils.job_prepare(job_backend)
        job_backend.progress(0, job_backend.job['config']['epochs'])

        logger.info("Start job")
        keras_model_utils.job_start(job_backend, trainer, keras_logger)

        job_backend.sync_weights()
        job_backend.done()

        logger.info("Done.")
        sys.exit(0)
    except KeyboardInterrupt:
        if job_backend.running:
            job_backend.set_status('ABORTED')

            if trainer.model:
                trainer.model.stop_training = True

            job_backend.sync_weights()
            job_backend.abort()
            logger.warning("Aborted.")

        sys.exit(1)
    except Exception as e:
        if trainer.model:
            trainer.model.stop_training = True

        logging.error(traceback.format_exc())
        job_backend.crash(e)
        raise



from __future__ import absolute_import

import sys
import time
import datetime
import six
from threading import Timer, Thread


class GeneralLogger(object):
    def __init__(self, job_backend, logger=None, error=False):
        self.error = error
        self.job_backend = job_backend
        self.buffer = ''
        self.last_timer = None
        self.last_messages = ''
        self.logger = logger or (sys.__stdout__ if error is False else sys.__stderr__)

    def get_time(self):
        ts = time.time()
        st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')

        return st + '.' + str(ts % 1)[2:6]

    def get_line(self, line):
        if line:
            return "[%s] %s\n" % (self.get_time(), line)

        return line

    def fileno(self):
        return self.logger.fileno()

    def flush(self):
        self.logger.flush()
        self.send_buffer()

    def send_buffer(self):
        self.last_timer = None

        if self.buffer:
            if self.job_backend:
                self.job_backend.write_log(self.buffer)

        self.buffer = ''

    def attach(self, buffer):
        """
        Read buffer until end (read() returns '') and sends it to self.logger and self.job_backend.

        :param buffer: a buffer instance with block read() method
        """
        def reader():
            while True:

                # read() needs to block
                buf = buffer.read(1)
                if buf == '':
                    return

                self.write(buf.decode("utf-8"))

        thread = Thread(target=reader)
        thread.daemon = True
        thread.start()

    def write(self, message):
        try:
            message = six.text_type(message)
        except:
            pass

        try:
            self.logger.write(message)

            self.last_messages += message
            if len(self.last_messages) > 20 * 1024:
                self.last_messages = self.last_messages[-20 * 1024:]

            for char in message:
                if '\b' == char or '\r' == char:
                    self.buffer = self.buffer[:-1]
                else:
                    self.buffer += char
        except:
            self.last_messages = ''
            self.buffer = ''
            pass

        if not self.last_timer:
            self.last_timer = Timer(1.0, self.send_buffer)
            self.last_timer.start()


import subprocess

import re


def get_branches():
    branches = subprocess.check_output(['git', 'branch']).decode("utf-8").strip().split('\n')

    return [x.strip(' *') for x in branches]


def get_current_branch():
    branches = subprocess.check_output(['git', 'branch']).decode("utf-8")
    m = re.search('\* ([^\s]+)', branches)
    current_branch = m.group(1) if m else None

    return current_branch


def get_current_commit_hash():
    commit_sha = subprocess.check_output(['git', 'rev-parse', 'HEAD']).decode("utf-8")

    return commit_sha.strip()


from __future__ import division
from __future__ import absolute_import

import os
import time
import datetime
import numpy as np
import six
import yaml


def get_option(dict, key, default=None, type=None):
    if key not in dict or dict[key] == '':
        return default

    if type == 'bool':
        bool(dict[key])

    return dict[key]


def unpack_full_job_id(full_id):
    if full_id.count('/') < 2:
        raise Exception('Not a valid full job id: ' + full_id)

    owner = full_id[0:full_id.index('/')]
    model = full_id[full_id.index('/') + 1:full_id.rindex('/')]
    id = full_id[full_id.rindex('/')+1:]

    return [owner, model, id]


def read_home_config(path = '~/.aetros.yml', logger=None):
    path = os.path.expanduser(path)
    custom_config = {}

    if os.path.exists(path):
        f = open(path, 'r')
        try:
            logger and logger.debug('Home config loaded from ' + os.path.realpath(path))
            custom_config = yaml.load(f)
        except:
            raise Exception('Could not load aetros home config at ' + os.path.realpath(path))

        if custom_config is None:
            custom_config = {}

    config = {
        'host': os.getenv('API_HOST') or 'trainer.aetros.com',
        'storage_dir': '~/.aetros',
        'ssh_key': os.getenv('AETROS_SSH_KEY'),
        'ssh': 'ssh',
        'ssl_verify': True,
    }

    config.update(custom_config)

    config['storage_dir'] = os.path.abspath(os.path.expanduser(config['storage_dir']))

    return config


def read_config(path = '.aetros.yml', logger=None):
    path = os.path.expanduser(path)
    home_config = read_home_config(logger=logger)

    config = home_config
    if os.path.exists(path):
        f = open(path, 'r')

        custom_config = yaml.load(f)
        if custom_config is None:
            custom_config = {}

        config.update(custom_config)

        logger and logger.debug('Config loaded from ' + os.path.realpath(path))

    if 'parameters' not in config:
        config['parameters'] = {}

    config['storage_dir'] = os.path.expanduser(config['storage_dir'])

    return config

def invalid_json_values(obj):
    if isinstance(obj, np.generic):
        return obj.item()
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, bytes):
        return obj.decode('cp437')

    if isinstance(map, type) and isinstance(obj, map):
        # python 3 map
        return list(obj)

    raise TypeError('Invalid data type passed to json encoder: ' + type(obj).__name__)


def get_time(self):
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')

    return st


def get_time_with_milli(self):
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')

    return st + '.' + str(ts % 1)[2:6]


def array_to_img(x, scale=True):
    """
    x should be shape (channels, width, height)
    """
    from PIL import Image
    if x.ndim != 3:
        raise Exception('Unsupported shape : ', str(x.shape), '. Need (channels, width, height)')
    if scale:
        x += max(-np.min(x), 0)
        x /= np.max(x)
        x *= 255
    if x.shape[0] == 3:
        # RGB
        if x.dtype != 'uint8':
            x = x.astype('uint8')
        return Image.fromarray(x.astype('uint8'), 'RGB')
    elif x.shape[0] == 1:
        # grayscale
        if x.dtype != 'uint8':
            x = x.astype('uint8')
        return Image.fromarray(x.reshape(x.shape[1], x.shape[2]), 'L')
    else:
        raise Exception('Unsupported channel number: ', x.shape[0])



from __future__ import absolute_import
from __future__ import print_function
import argparse
import json
import os
import sys
import urllib
from threading import Lock
import time


class PredictionServerCommand:

    model = None
    job_model = None
    def __init__(self, logger):
        self.logger = logger

    def main(self, args):
        import aetros.const

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter, prog=aetros.const.__prog__ + ' server')
        parser.add_argument('id', nargs='?', help='job id')
        parser.add_argument('--weights', help="Weights path. Per default we try to find it in the ./weights/ folder or download it.")
        parser.add_argument('--latest', action="store_true", help="Instead of best epoch we upload latest weights.")
        parser.add_argument('--tf', action='store_true', help="Uses TensorFlow instead of Theano")
        parser.add_argument('--port', help="Changes port. Default 8000")
        parser.add_argument('--host', help="Changes host. Default 127.0.0.1")

        parsed_args = parser.parse_args(args)
        self.lock = Lock()

        sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)

        if not parsed_args.id:
            parser.print_help()
            sys.exit()

        os.environ['KERAS_BACKEND'] = 'tensorflow' if parsed_args.tf else 'theano'

        self.model = self.start_model(parsed_args)
        self.start_webserver('127.0.0.1' if not parsed_args.host else parsed_args.host, 8000 if not parsed_args.port else int(parsed_args.port))

    def start_model(self, parsed_args):
        from aetros import keras_model_utils
        from aetros.backend import JobBackend
        from aetros.logger import GeneralLogger
        from aetros.Trainer import Trainer
        from aetros.keras_model_utils import ensure_dir

        if not parsed_args.id:
            print("No job id given.")
            sys.exit(1)

        print("...")
        self.lock.acquire()
        job_backend = JobBackend(parsed_args.id)
        job_backend.load_light_job()
        self.job_model = job_backend.get_job_model()

        if parsed_args.weights:
            weights_path = parsed_args.weights
        elif parsed_args.latest:
            weights_path = self.job_model.get_weights_filepath_latest()
        else:
            weights_path = self.job_model.get_weights_filepath_best()

        print("Check weights ...")

        if not os.path.exists(weights_path) or os.path.getsize(weights_path) == 0:
            weight_url = job_backend.get_best_weight_url(parsed_args.id)
            if not weight_url:
                print("No weights available for this job.")
                exit(1)

            print(("Download weights %s to %s .." % (weight_url, weights_path)))
            ensure_dir(os.path.dirname(weights_path))

            f = open(weights_path, 'wb')
            f.write(urllib.urlopen(weight_url).read())
            f.close()

        keras_model_utils.job_prepare(job_backend)

        trainer = Trainer(job_backend)

        self.job_model.set_input_shape(trainer)

        print("Loading model ...")
        model = self.job_model.get_built_model(trainer)

        print(("Load weights %s ..." % (weights_path,)))
        self.job_model.load_weights(model, weights_path)
        print("Locked and loaded.")

        self.lock.release()

        return model

    def start_webserver(self, host, port):
        import cherrypy
        import numpy

        class WebServer(object):
            def __init__(self, lock, job_model, model):
                self.job_model = job_model
                self.model = model
                self.lock = lock

            @cherrypy.expose
            def predict(self, path=None, paths=None, uploads=None, inputs=None):
                self.lock.acquire()
                result = {'times': {}}

                try:
                    start = time.time()

                    if not path and not paths and not uploads and not inputs:
                        return json.dumps({'error': 'not_input_given'})

                    if path and not paths and not uploads and not inputs:
                        paths = [path]

                    encoded_inputs = []
                    if paths:
                        if not isinstance(paths, list):
                            paths = [paths]

                        for idx, file_path in enumerate(paths):
                            encoded_inputs.append(self.job_model.convert_file_to_input_node(file_path, self.job_model.get_input_node(idx)))

                    if inputs:
                        if not isinstance(inputs, list):
                            inputs = [inputs]

                        for idx, input in enumerate(inputs):
                            encoded_inputs.append(self.job_model.encode_input_to_input_node(input, self.model.input_layers[idx]))

                    if uploads:
                        if not isinstance(uploads, list):
                            uploads = [uploads]

                        for idx, upload in enumerate(uploads):
                            encoded_inputs.append(self.job_model.convert_file_to_input_node(upload.filename, self.job_model.get_input_node(idx)))

                    result['times']['prepare_fetch_input'] = time.time() - start

                    start = time.time()
                    prediction = self.job_model.predict(self.model, numpy.array(encoded_inputs))
                    result['times']['prediction'] = time.time() - start

                    self.lock.release()
                except Exception as e:
                    self.lock.release()
                    return json.dumps({'error': type(e).__name__, 'message': e.message})

                result['prediction'] = prediction

                return json.dumps(result)

        cherrypy.config.update({
            'server.socket_host': host,
            'server.socket_port': port,
            'server.thread_pool': 1,
        })

        print("Starting server ... Use http://127.0.0.1:8000/predict?paths=path_to_image.jpg or upload files named 'uploads[0]' to http://127.0.0.1:8000/predict.")
        cherrypy.quickstart(WebServer(self.lock, self.job_model, self.model))



from __future__ import absolute_import
import argparse
import os
import sys
import aetros.const

class PredictCommand:

    def __init__(self, logger):
        self.logger = logger

    def main(self, args):

        from aetros.predict import predict
        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter, prog=aetros.const.__prog__ + ' predict')
        parser.add_argument('job', nargs='?', help='the job id, e.g. GoVDO1Njm')
        parser.add_argument('--api-key', help="Secure key. Alternatively use API_KEY environment varibale.")
        parser.add_argument('--insights', action='store_true', help="activates insights")
        parser.add_argument('--weights', help="Weights path. Per default we try to find it in the ./weights/ folder.")
        parser.add_argument('-i', nargs='+', help="Input (path or url). Multiple allowed")
        parser.add_argument('--tf', action='store_true', help="Uses TensorFlow instead of Theano")

        parsed_args = parser.parse_args(args)

        if not parsed_args.job:
            parser.print_help()
            sys.exit()

        os.environ['KERAS_BACKEND'] = 'tensorflow' if parsed_args.tf else 'theano'
        predict(parsed_args.job, parsed_args.i, parsed_args.insights, parsed_args.weights, api_key=parsed_args.api_key)



from setuptools import setup
from setuptools import find_packages
import aetros.const

setup(name='aetros',
      version=aetros.const.__version__,
      description='Python SDK for aetros.com',
      author='Marc J. Schmidt',
      author_email='marc@marcjschmidt.de',
      url='https://github.com/aetros/aetros-cli',
      download_url='https://github.com/aetros/aetros-cli/tarball/' + aetros.const.__version__,
      license='MIT',
      entry_points={
          'console_scripts': ['aetros=aetros:main'],
      },
      install_requires=('keras',
                        'requests',
                        'numpy',
                        'scipy',
                        'h5py',
                        'coloredlogs',
                        'psutil>=5.2.0',
                        'pyyaml>=3.0',
                        'cherrypy>=7.1.0',
                        'Pillow>=4.0.0',
                        'py-cpuinfo==0.2.3',
                        'py-cpuinfo==0.2.3',
                        'msgpack-python>=0.4.8'),
      packages=find_packages())



from __future__ import absolute_import
from __future__ import print_function
import argparse
import os
import sys
import time
import yaml

from aetros.backend import BackendClient, EventListener, ApiClient


class RunClient(BackendClient):
    def __init__(self, api_host, api_port, event_listener):
        BackendClient.__init__(self, api_host, api_port, event_listener)
        self.model_id = None

    def configure(self, model_id, api_key):
        self.model_id = model_id
        self.api_key = api_key

    def on_connect(self, reconnect=False):
        self.send_message({'register_run': self.api_key, 'model': self.model_id})
        messages = self.wait_for_at_least_one_message(self.s)

        if not messages:
            return False

        message = messages.pop(0)
        if isinstance(message, dict) and 'a' in message:
            if "ACCESS_DENIED" in message['a']:
                print("Access denied. Exiting")
                self.event_listener.fire('stop')
                self.active = False
                return False

            if "REGISTERED" in message['a']:
                self.registered = True
                print("Connected to %s:%d." % (self.api_host, self.api_port))
                self.event_listener.fire('registration')
                self.handle_messages(messages)
                return True

        print("Registration of job %s failed." % (self.job_id,))
        return False


class RunCommand:
    def __init__(self, logger):
        self.logger = logger
        self.client = None
        self.registered = False
        self.active = True

    def main(self, args):
        import aetros.const

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                         prog=aetros.const.__prog__ + ' run')
        parser.add_argument('--server', help="On which AETROS server the command should be executed. Default read in .aetros.yml")
        parser.add_argument('--model', help="Under which model this job should be listed. Default read in .aetros.yml")
        parser.add_argument('--host', help="Default trainer.aetros.com. Read from environment variable API_HOST.")
        parser.add_argument('--port', help="Default 8051. Read from environment variable API_PORT.")
        parser.add_argument('--config', help="Default .aetros.yml in current working directory.")
        parser.add_argument('--api-key', help="Read from environment variable API_KEY.")

        parsed_args = parser.parse_args(args)

        if not parsed_args.secure_key:
            parser.print_help()
            sys.exit()

        if parsed_args.max_jobs:
            self.max_parallel_jobs = int(parsed_args.max_jobs)
        if parsed_args.show_stdout:
            self.show_stdout = True

        config_path = parsed_args.config if parsed_args.config else '.aetros.yml'
        config = {}

        if os.path.exists(config_path):
            with open(config_path, 'r') as file:
                config = yaml.load(file.read())

        event_listener = EventListener()

        api_host = os.getenv('API_HOST')
        api_port = os.getenv('API_PORT')
        api_key = os.getenv('API_KEY')
        model_id = None

        if config['api_key']:
            api_key = config['api_key']
        if parsed_args.api_key:
            api_key = parsed_args.api_key

        if config['host']:
            api_host = config['host']
        if parsed_args.host:
            api_host = parsed_args.host

        if config['port']:
            api_port = config['host']
        if parsed_args.host:
            api_port = parsed_args.port

        if config['model']:
            model_id = config['model']
        if parsed_args.model:
            model_id = parsed_args.model

        event_listener.on('registration', self.on_registration)
        event_listener.on('disconnect', self.on_disconnect)
        event_listener.on('stop', self.on_stop)

        self.client = RunClient(api_host, api_port, event_listener)
        self.client.configure(model_id, api_key)
        self.client.start()

        while self.active:
            if self.registered:
                pass

            time.sleep(1)

    def on_registration(self, params):
        self.registered = True

    def on_disconnect(self, params):
        self.registered = False

    def on_stop(self, params):
        self.active = False



from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from threading import Lock

def is_generator(obj):
    import inspect

    return obj is not None and (inspect.isgeneratorfunction(obj) or inspect.isgenerator(obj) or hasattr(obj, 'next') or hasattr(obj, '__next__'))


class Trainer():
    """
    :type job_backend : aetros.backend.JobBackend
    :type settings : dict
    """

    def __init__(self, job_backend):
        self.job_backend = job_backend
        self.input_shape = []

        # training sample count per epoch for generator. same name as in keras fit_generator
        self.samples_per_epoch = None #used when simple code uses a generator
        # validation sample count per epoch for generator. same name as in keras fit_generator
        self.nb_val_samples = None #used when simple code uses a generator
        self.nb_val_steps = None

        self.callbacks = [] #used by simple models
        self.classes = None #set by auto_dataset
        self.output_size = None #for code generator output layer, set by auto_dataset
        self.model = None

        self.job_model = job_backend.get_job_model()
        self.settings = {}

        if 'batchSize' in self.job_backend.job['config']:
            self.settings['batchSize'] = self.job_backend.job['config']['batchSize']

        if 'epochs' in self.job_backend.job['config']:
            self.settings['epochs'] = self.job_backend.job['config']['epochs']

        self.lock = Lock()

    @property
    def logger(self):
        return self.job_backend.logger

    def set_model(self, model):
        self.model = model

    def get_batch_size(self):
        return self.job_backend.get_job_model().get_batch_size()

    def set_generator_validation_nb(self, number):
        """
        sets self.nb_val_samples which is used in model.fit if input is a generator
        :param number:
        :return:
        """

        self.nb_val_samples = number
        diff_to_batch = number % self.get_batch_size()
        if diff_to_batch > 0:
            self.nb_val_samples += self.get_batch_size() - diff_to_batch

        import keras
        if '1' != keras.__version__[0]:
            self.nb_val_samples = self.nb_val_samples // self.get_batch_size()

    def set_generator_training_nb(self, number):
        """
        sets self.samples_per_epoch which is used in model.fit if input is a generator
        :param number:
        :return:
        """

        self.samples_per_epoch = number
        diff_to_batch = number % self.get_batch_size()
        if diff_to_batch > 0:
            self.samples_per_epoch += self.get_batch_size() - diff_to_batch

    def set_status(self, status):
        self.job_backend.set_status(status)

    def set_info(self, name, value):
        self.job_backend.set_info(name, value)

    def has_generator(self, dict):
        for v in dict.values():
            if is_generator(v):
                return True

        return False

    def get_first_generator(self, dict):
        for v in dict.values():
            if is_generator(v):
                return v

        return None

    def set_job_system_info(self, key, value):
        self.job_backend.set_system_info(key, value)

    def set_job_info(self, key, value):
        self.job_backend.set_info(key, value)



import unittest

import sys

import pytest

from aetros.KerasCallback import KerasCallback
from aetros.Trainer import is_generator
from aetros.backend import JobBackend

class TestAetrosCallback(unittest.TestCase):

    def test_set_validation_data(self):
        job_backend = JobBackend('test')
        job_backend.job = {'id': 'test', 'index': 1, 'modelId': 'test/model'}

        keras_callback = KerasCallback(job_backend, sys.stdout)

        keras_callback.set_validation_data(([1, 2, 3], [1, 2, 3]))
        self.assertEqual(keras_callback.data_validation_size, 3)

        with pytest.raises(Exception):
            keras_callback.set_validation_data([[1, 2], [1, 2]])

        # aetros format
        keras_callback.set_validation_data({'x': {'input': [1, 2, 3]}, 'y': {'output': [1, 2, 3]}})
        self.assertEqual(keras_callback.data_validation_size, 3)

        keras_callback.set_validation_data({'input': [1, 2, 3], 'output': [1, 2, 3]})
        self.assertEqual(keras_callback.data_validation_size, 3)

        keras_callback.set_validation_data(([1, 2, 3], [1, 2, 3]), 5)
        self.assertEqual(keras_callback.data_validation_size, 3)

        def generator():
            yield ([1, 2], [1, 2])

        self.assertTrue(is_generator(generator))

        with pytest.raises(Exception):
            keras_callback.set_validation_data(generator)

        with pytest.raises(Exception):
            keras_callback.set_validation_data((generator, generator))

        with pytest.raises(Exception):
            keras_callback.set_validation_data([generator, generator], 6)

        keras_callback.set_validation_data(generator, 5)
        self.assertEqual(keras_callback.data_validation_size, 5)

        keras_callback.set_validation_data((generator, generator), 4)
        self.assertEqual(keras_callback.data_validation_size, 4)



from __future__ import absolute_import
from __future__ import print_function
import argparse
import os


class UploadWeightsCommand:

    def __init__(self, logger):
        self.logger = logger

    def main(self, args):

        from aetros import keras_model_utils

        import aetros.const
        from aetros.backend import JobBackend
        from aetros.Trainer import Trainer

        parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter, prog=aetros.const.__prog__ + ' upload-weights')
        parser.add_argument('id', help='model name or job id')
        parser.add_argument('weights', help="Weights path")
        parser.add_argument('--api-key', help="Secure key. Alternatively use API_KEY environment variable.")
        parser.add_argument('--kpi', help="You can overwrite or set the KPI for this job")
        parser.add_argument('--latest', action="store_true", help="Instead of best epoch we upload latest weights.")
        parsed_args = parser.parse_args(args)

        if not parsed_args.id or not parsed_args.weights:
            parser.print_help()
            return

        job_backend = JobBackend(api_key=parsed_args.api_key)

        if '/' in parsed_args.id and '@' not in parsed_args.id:
            job_backend.create(parsed_args.id)

        job_backend.load(parsed_args.id)

        if job_backend.job is None:
            raise Exception("Job not found")

        weights_path = parsed_args.weights

        if not os.path.exists(weights_path):
            raise Exception('Weights file does not exist in ' + weights_path)

        print("Uploading weights to %s of %s ..." % (job_backend.job_id, job_backend.model_id))

        job_backend.upload_weights('weights.hdf5', weights_path, float(parsed_args.kpi) if parsed_args.kpi else None)
